#! /usr/bin/env perl

package Smash::Analyses::Comparative::16S;
use strict;
use warnings;
use Smash::Core;
use Smash::Global qw(:all);
use Smash::Utils::Taxonomy qw(:all);
use Smash::Utils::MatrixIO qw(read_R_matrix write_R_matrix);

our @ISA = qw(Smash::Analyses::Comparative);

my $READ_PHYLO_MAP;

sub init {
	my $this = shift;
	$this->SUPER::init();
	$this->{TYPE} = "16S";
	Smash::Utils::Taxonomy::init();
	$this->{READ_LENGTH_THRESHOLD} = 250 unless $this->{READ_LENGTH_THRESHOLD};
	$this->{CONFIDENCE_THRESHOLD}  = 0.6 unless $this->{CONFIDENCE_THRESHOLD};
	if ($this->experiment eq "16S") {
		$this->{READ_LENGTH_THRESHOLD} = 250;
		$this->{CONFIDENCE_THRESHOLD}  = 0.5;
	}
}

sub attach_sample {
	my $this       = shift;
	my $metagenome = shift || die "attach_sample needs metagenome";
	my $label      = shift || die "attach_sample needs label";
	my $smash      = $this->smash;

	$smash->finish() if $smash;
	$smash = new Smash::Core(METAGENOME => $metagenome);
	$smash->init();
	$this->{SMASH} = $smash;

	$this->{LABEL} = $label;
	push(@{$this->{LABELS}}, $label);
}

sub get_data_count {
	my $this  = shift;
	my $metagenome = shift;
	my $dbh   = $this->smash->get_db_handle();
	my $sth   = $dbh->prepare_cached("SELECT COUNT(DISTINCT template_id) FROM readinfo r INNER JOIN library l USING (library_id) INNER JOIN sample s USING (sample_id) WHERE metagenome_id=? AND r.length >= ?");
	$sth->execute($metagenome, 100);
	my ($count) = $sth->fetchrow_array();
	$sth->fetchrow_array();
	return $count;
}

########################
# Get the NCBI taxonomy information
########################

sub analyze_one_sample {

	my $this = shift;
	my $label = $this->label;
	my $smash = $this->smash;
	my $feature_count = $this->feature_count;
	my $metagenome = $smash->metagenome;
	my $class_file = $smash->analyses_dir($metagenome)."/$metagenome.16S.@{[$this->refdb]}.class";

	my $required_level = $Rank2Num{$this->level};

	my %Template;
	my %ReadLength;
	my $dbh = $smash->get_db_handle;
	{
		my $sth = $dbh->prepare("SELECT read_id, template_id, length FROM readinfo WHERE read_id LIKE ?");
		$sth->execute("$metagenome.%");
		my ($id, $template, $length);
		$sth->bind_columns(\$id, \$template, \$length);
		while ($sth->fetch()) {
			$Template{$id} = $template;
			$ReadLength{$id} = $length;
		}
	}

	my $db = new File::Temp();
	my $dbname = $db->filename;
	
	$dbh = DBI->connect("DBI:SQLite:dbname=$dbname.sqlite", "", "", {AutoCommit => 0, RaiseError => 1}) || die "Couldn't connect to database: " .DBI->errstr;

	$dbh->do("CREATE TABLE templates(template_id VARCHAR(255) NOT NULL, taxonomy_id INTEGER NOT NULL)");
	$dbh->do("CREATE TABLE templates_filtered(template_id VARCHAR(255) NOT NULL, taxonomy_id INTEGER NOT NULL)");
	$dbh->do("CREATE TABLE template_occurrences(template_id VARCHAR(255) PRIMARY KEY, occurrence INTEGER NOT NULL)");
	$dbh->do("CREATE TABLE template_map(template_id VARCHAR(255) NOT NULL, taxonomy_id INT NOT NULL, count FLOAT(6,4))");
	#$dbh->do("CREATE TABLE taxonomy_map(taxonomy_id INT NOT NULL, count FLOAT(12,4))");

	{ # to keep $sth within limited scope
		my $sth = $dbh->prepare("INSERT INTO templates(template_id, taxonomy_id) VALUES(?, ?)");
		open(CLASS, "<$class_file") || die "Cannot open $class_file: $!";
		LINE:while (<CLASS>) {
			chomp();

			my @pairs  = split(/\t/);
			my $read   = shift(@pairs);
			shift(@pairs); # dont need strand now
			next LINE if ($ReadLength{$read} < $this->{READ_LENGTH_THRESHOLD});

			my $phylum;
			my $tax_id;
			my $prev_name;
			my $saved_tax_id = undef;
			RANK: while (scalar(@pairs) > 1) {
				my $name = shift(@pairs); 
				my $rank = shift(@pairs); 
				my $prob = shift(@pairs);

				# If the probability is too low, we quit!

				if ($prob < $this->{CONFIDENCE_THRESHOLD}) { last RANK; }

				# Fix names: complete Incertae Sedis, and remove quotes

				$name   =~ s/^['"]//;
				$name   =~ s/['"]$//;
				if ($name =~ /^Incertae/i) {
					$name = "$prev_name $name";
				}
				$tax_id = $BergeyRankedName2BergeyTax{$rank}{$name};

				my $level = $Rank2Num{$rank};

				# Save the tax_id if we reach the level we want
				# If we passed it without seeing what we want, try
				# going back up until we pass it the other way

				if ($level == $required_level) { # right there!
					$saved_tax_id = $tax_id;
				} elsif ($level > $required_level && !defined($saved_tax_id)) { # passed it without saving!
					my $child = $tax_id;
					PARENT: while (1) {
						my $parent       = $BergeyTax2BergeyParent{$child};
						my $parent_level = $Rank2Num{$BergeyTax2BergeyRank{$parent}};
						if ($parent_level < $required_level) {
							$saved_tax_id = $child;
							last PARENT;
						}
						$child = $parent;
					}
				}
				$prev_name = $name;
			}

			# At this point, we have $tax_id = the final prediction, $saved_tax_id = prediction closest to the required level

			if (!defined($saved_tax_id)) { # didnt reach the required level
				$saved_tax_id = $tax_id;
			}

			my $rank = $BergeyTax2BergeyRank{$saved_tax_id};
			my $name = $BergeyTax2BergeyName{$saved_tax_id};

			# $saved_tax_id now has the Bergey tax id for the right candidate at approximately $required_level

			if (uc($this->tree) eq "NCBI") {

				# If it is Incertae Sedis, or a subblahblah, go one level up
				# 17.06.2010-Mani: Why? I forgot why I did this!
				# Perhaps because you cannot get the NCBI tax id's for those guys!

				if ($name =~ /Incertae/i || $rank =~ /^sub/) {
					$saved_tax_id = $BergeyTax2BergeyParent{$saved_tax_id};
					$rank      = $BergeyTax2BergeyRank{$saved_tax_id};
					$name      = $BergeyTax2BergeyName{$saved_tax_id};
				}

				# convert to NCBI tax ids.

				my $ncbi_tax_id = bergey2ncbi_name_rank($name, $rank);
				if (defined($ncbi_tax_id)) {
					# Want at least $opt_level, others dont matter (buffer 4)
					# For genus, this is suborder, which already includes family!
					if ($Rank2Num{$rank} >= $required_level-4) {
						#print "$read => $rank=$name => $tax_id => $ncbi_tax_id\n";
						$sth->execute($Template{$read}, $ncbi_tax_id);
					}
				} else {
print "Skipping $name:$rank\n";
				}
			} else {
				# Want at least $opt_level, others dont matter (buffer 4)
				# For genus, this is suborder, which already includes family!
				if ($Rank2Num{$rank} >= $required_level-4) { 
					$sth->execute($Template{$read}, $tax_id);
				}
			}

			# Note the final mapping down

			if (1 == 0) { # disabled for now
				$rank = $BergeyTax2BergeyRank{$tax_id};
				if ($Rank2Num{$rank} >= $Rank2Num{"phylum"}) {
					$name = $BergeyTax2BergeyName{$tax_id};
					my $ncbi_tax_id = $NCBIRankedName2NCBITax{$rank}{$name} || $NCBIRankedName2NCBITax{$rank}{"$name ($rank)"} || $NCBIRankedName2NCBITax{"no rank"}{$name};
					if (defined($ncbi_tax_id)) {
						print $READ_PHYLO_MAP "$read\t$ncbi_tax_id\n";
					} else {
						print $READ_PHYLO_MAP "$read\t-$tax_id\n";
					}
				}
			}
		}
		close(CLASS);
		#$sth->finish();
	}

	$dbh->do("INSERT INTO templates_filtered(template_id, taxonomy_id) SELECT DISTINCT template_id, taxonomy_id FROM templates");
	$dbh->do("INSERT INTO template_occurrences(template_id, occurrence) SELECT template_id, COUNT(*) AS occurrence FROM templates GROUP BY template_id");
	$dbh->do("INSERT INTO template_map(template_id, taxonomy_id, count) SELECT template_id, taxonomy_id, 1.0/occurrence FROM templates INNER JOIN template_occurrences USING (template_id)");
	#$dbh->do("CREATE TABLE taxonomy_map AS SELECT taxonomy_id, SUM(count) AS count FROM template_map GROUP BY taxonomy_id");
	$dbh->commit();

	my $assigned = 0;
	{
		my ($tax, $count);
		my $sth = $dbh->prepare("SELECT taxonomy_id, SUM(count) FROM template_map GROUP BY taxonomy_id");
		$sth->execute();
		$sth->bind_columns(\$tax, \$count);
		while ($sth->fetch()) {
			$feature_count->{$tax}->{$label} = $count;
			$assigned += $count;
		}
	}
	#$dbh->do("DROP TABLE taxonomy_map");

	$dbh->do("DROP TABLE template_map");
	$dbh->do("DROP TABLE template_occurrences");
	$dbh->do("DROP TABLE templates_filtered");
	$dbh->do("DROP TABLE templates");
	$dbh->commit();
	$dbh->disconnect();
	undef $dbh;
	unlink "$dbname.sqlite";

	return $assigned;
}

sub get_attribute {
	my ($string, $attribute) = @_;
	my (undef, $value) = $string =~ m#${attribute}=(["'])([^\1]*?)\1#i;
	return $value;
}

sub convert_to_abundance {
	my $this = shift;
	my $feature_count = shift;

	my @samples = @{$this->labels};
	my $data_count = $this->data_count;
	my ($rdp_classifier_dir) = $this->smash->software_dir("rdp_classifier", "current");

	# which lineage finder to use?

	my $lineage_finder;
	my %RankedName2Tax;
	my %Tax2Name;
	if (uc($this->tree) eq "NCBI") {
		$lineage_finder = \&get_full_ncbi_lineage;
		%RankedName2Tax = %NCBIRankedName2NCBITax;
		%Tax2Name = %NCBITax2NCBIName;
	} else {
		$lineage_finder = \&get_full_bergey_lineage;
		%RankedName2Tax = %BergeyRankedName2BergeyTax;
		%Tax2Name = %BergeyTax2BergeyName;
	}
	# normalize 16S counts by copy number, now we have ~number of individuals

	my %CopyNumber;
	my %Count;

	open(COPYNUMBER, "<$rdp_classifier_dir/rrn_copy_numbers.txt");
	while (<COPYNUMBER>) {
		if (!m/^#/) {
			my ($genus, $copies16S) = (split(/\t/))[0,3];
			my $tax_id = $RankedName2Tax{"genus"}{$genus};
			if ($copies16S !~ /NA/ && $tax_id) {
				my @lineage = $lineage_finder->($tax_id);
				foreach my $ancestor (@lineage) {
					$CopyNumber{$ancestor} = 0;
					$Count{$ancestor}      = 0;
				}
			}
		}
	}
	seek COPYNUMBER, 0, 0;
	while (<COPYNUMBER>) {
		if (!m/^#/) {
			my ($genus, $copies16S) = (split(/\t/))[0,3];
			my $tax_id = $RankedName2Tax{"genus"}{$genus};
			if ($copies16S !~ /NA/ && $tax_id) {
				my @lineage = $lineage_finder->($tax_id);
				foreach my $ancestor (@lineage) {
					$CopyNumber{$ancestor} += $copies16S;
					$Count{$ancestor}++;
				}
			}
		}
	}
	close(COPYNUMBER);

	# Calculate average of the whole phylo tree

	foreach my $key (keys %CopyNumber) {
		$CopyNumber{$key} /= $Count{$key};
		#print "$key\t$Tax2Name{$key}\t$Tax2Rank{$key}\t$CopyNumber{$key}\n";
	}

	# record raw count in case of replicate

	if (defined($this->replicate)) {
		my $file_name = sprintf("%s.feature_vector.rawcount.%s.%s.%s", $this->name, $this->type, $this->refdb, $this->replicate);
		write_R_matrix($file_name, $feature_count);
	}

	# As of 28.09.2009, rrnDB has copy number info for 979 genomes.
	# As of 17.06.2010, rrnDB has 1068 genomes.

	# 1. Normalize for 16S copy number variation
	# 2. Normalize for data size
	# Average copy number is 3.96; average genome size for all refgenomes we have is 3539003
	# Thus you expect 3.96 copies per 3539003 bp sampled, roughly
	# For $sample_size bp, we then expect 3.96*$sample_size/3539003 16S molecules
	# This is the data normalization. But we dont do it now, we just normalize on total 16S sequences generated.
	# EVERYTHING SUMS TO 1 AFTER HERE


	my @features = keys %$feature_count;
	foreach my $label (@samples) {
		foreach my $tx (@features) {
			my $c = ($feature_count->{$tx}{$label} || 0);
			if ($c) {

				# Get the copy number at the lowest common ancestor
				# Unassigned or unknown 16S gets copy number of root!

				my $copy_number = $CopyNumber{1}; 
				if ($tx != -1) {
					my @lineage = $lineage_finder->($tx);
					LINEAGE: foreach my $ancestor (@lineage) {
						$copy_number = $CopyNumber{$ancestor};
						if ($copy_number) {
#if ($tx == 572511 || $tx == 3017) {
#	printf "Blautia: $copy_number from %s\n", $Tax2Name{$ancestor};
#}
							last LINEAGE;
						}
					}
				}
				$c = $c/$copy_number;
				$feature_count->{$tx}{$label} = $c;
			}
		}
	}

	return;
}

1;
