#! /usr/bin/env perl

package Smash::Analyses::Comparative::Functional;
use strict;
use warnings;
use Smash::Core;
use Smash::Utils::MatrixIO qw(write_R_matrix zero_fill_matrix);

our @ISA = qw(Smash::Analyses::Comparative);

sub normalize {shift->{NORMALIZE}}

sub new {
	my $class = shift;
	my %param = @_;
	my @available = qw(KEGG eggNOG);
	my ($name) = grep {lc($param{REF_DB}) eq lc($_)} @available;
	$class .= "::$name";
	my $this  = bless {%param}, $class;
	select(STDERR); $| = 1; select(STDOUT);
	$this->{PROGRESS} = \*STDERR;
	$this->{TYPE} = "functional";
	return $this;
}

sub init {
	my $this = shift;
	if ($this->normalize eq "data") {
		warn "functional annotation normally uses hit-normalization and not data-normalization.";
	}
	$this->SUPER::init();
}

sub attach_sample {
	my $this     = shift;
	my $genepred = shift || die "attach_sample needs genepred";
	my $label    = shift || die "attach_sample needs label";
	my $smash    = $this->smash;

	$smash->finish() if $smash;
	$smash = new Smash::Core(GENEPRED => $genepred);
	$smash->init();
	$this->{SMASH} = $smash;

	$this->{LABEL} = $label;
	push(@{$this->{LABELS}}, $label);
}

sub get_data_count {
	my $this  = shift;
	my $genepred = shift;
	my $dbh   = $this->smash->get_db_handle();
	my $sth   = $dbh->prepare_cached("SELECT gene_unit_count FROM summary_gene_stats WHERE gene_prediction_id=?");
	$sth->execute($genepred);
	my ($count) = $sth->fetchrow_array();
	$sth->fetchrow_array();
	return $count;
}

sub analyze_one_sample {
	my $this         = shift;
	my ($bootstrap_label) = @_;
	my $label        = $this->label;
	my $smash        = $this->smash;
	my $metagenome   = $smash->metagenome;
	my $genepred     = $smash->genepred;
	my $genepred_id  = $smash->get_id_by_name("gene_prediction", $genepred);
	my $genepred_dir = $smash->genepred_dir($genepred);

	my $kegg_map     = "$genepred_dir/$genepred.@{[$this->refdb]}mapping.txt";

	# hack

	#warn "This includes a hack. Please fix the code soon!\n";
	#$kegg_map     = "$genepred.@{[$this->refdb]}mapping.txt";

	# end hack

	my $ResampledDataCount;

=begin OLD_REPLICATE

	if (defined($this->replicate)) {
		my $boot_dir  = $smash->bootstrap_dir($metagenome)."/$bootstrap_label";
		my $list_file = "$boot_dir/replicate.".$this->replicate;
		open(LIST, "<$list_file") || die "Cannot open $list_file: $!";
		while (<LIST>) {
			chomp();
			if ($ResampledDataCount->{$_}) {
				$ResampledDataCount->{$_}++; 
			} else {
				$ResampledDataCount->{$_} = 1;
			}
		}
		close(LIST);
	}

=cut

	my $Coverage;
	my $Length;
	{
		my ($gene, $length, $cov);
		my $dbh = $smash->get_db_handle;
		my $sth = $dbh->prepare("SELECT gene_name, coverage FROM gene_details WHERE gene_prediction_id=?");
		$sth->execute($genepred_id);
		$sth->bind_columns(\$gene, \$cov);
		while ($sth->fetch()) {
			$Coverage->{$gene} = $cov;
		}
		$sth = $dbh->prepare("SELECT external_id, length FROM gene WHERE gene_prediction_id=?");
		$sth->execute($genepred_id);
		$sth->bind_columns(\$gene, \$length);
		while ($sth->fetch()) {
			$Length->{$gene} = $length;
		}
	}

	$this->analyze_file($kegg_map, $Coverage, $Length, $ResampledDataCount);
}

sub analyze_one_genome {
	my $this         = shift;
	my $label        = $this->label;
	my $smash        = $this->smash;
	my $kegg_map     = "$label.@{[$this->refdb]}mapping.txt";

	my $Coverage = {};
	my $Length = {};
	{
		my ($gene, $length, $cov);
		my $dbh = $smash->get_refgenomedb_handle;
		my $sth = $dbh->prepare("SELECT gene.external_id, gene.length FROM gene WHERE external_id like ?");
		$sth->execute("$label.%");
		$sth->bind_columns(\$gene, \$length);
		while ($sth->fetch()) {
			$Length->{$gene} = $length;
			$Coverage->{$gene} = 1;
		}
	}
	$this->analyze_file($kegg_map, $Coverage, $Length);
}

sub convert_to_abundance {
	my $this          = shift;
	my $feature_count = shift;

	zero_fill_matrix($feature_count);

	# record raw count in case of replicate

	if (defined($this->replicate)) {
		my $file_name = sprintf("%s.feature_vector.rawcount.%s.%s.%s", $this->name, $this->type, $this->refdb, $this->replicate);
		write_R_matrix($file_name, $feature_count);
	}
}

sub get_next_uncommented_line {
	my $this = shift;
	my $fh = shift;
	my $line;

	return 0 if eof $fh;

	# get the next non-comment line

	do {
		$line = <$fh>;
	} while ($line =~ /^#/);

	return 0 unless $line;
	return 0 if ($line =~ /^#/);
	chomp($line);
	return $line;
}

1;

package Smash::Analyses::Comparative::Functional::KEGG;
use strict;
use warnings;
use Smash::Utils::MatrixIO qw(read_R_matrix write_R_matrix transpose_matrix write_two_column_hash);

our @ISA = qw(Smash::Analyses::Comparative::Functional);

sub ko2gene                {shift->{KO2GENE}}
sub gene_physical_coverage {shift->{GENE_PHYS_COVERAGE}}
sub size_normalize         {shift->{SIZE_NORMALIZE}}

sub init {
	my $this = shift;
	$this->{SIZE_NORMALIZE} = 0;
	$this->SUPER::init();
}

sub analyze_file_1 {
	my $this         = shift;
	my $label        = $this->label;
	my $FeatureCount = $this->feature_count;
	my $genepred     = $this->smash->genepred;
	my $gp_id        = $this->smash->get_id_by_name("gene_prediction", $genepred);
	my ($kegg_map, $Coverage, $Length, $ResampledDataCount) = @_;

	# Read the kegg_maps

	open(KEGG, "<$kegg_map") || die "Cannot open $kegg_map: $!";

	# Process the whole dataset, coverage needs to be used here, since this is the original data

	my $dbh = $this->smash->get_db_handle;
	my $sth = $dbh->prepare("SELECT contig_id, start, end, strand FROM gene WHERE external_id=?");

	my $prev_gene = "dummy007";
	my ($contig, $cstart, $cend, $strand);
	LINE:while (my ($gene, $kegg, $gstart, $gend, $kstart, $kend) = $this->parse_next_kegg_map_line(\*KEGG)) {

		# get full gene coordinates
		if ($gene ne $prev_gene) {
			$sth->execute($gene);
			($contig, $cstart, $cend, $strand) = $sth->fetchrow_array();
		}

		# get virtual DNA alignment positions from protein alignment positions
		# both are one based, closed interval, as always in BLAST outputs

		($gstart, $kstart) = map {($_-1)*3+1} ($gstart, $kstart);
		($gend, $kend)     = map {$_*3} ($gend, $kend);

		# convert alignment gene coordinates to contig coordinates

		if ($strand eq "+") {
			($gstart, $gend) = map {$cstart + $_ - 1} ($gstart, $gend);
		} else {
			($gstart, $gend) = map {$cend - $_ + 1} ($gend, $gstart);
		}

		# convert protein2kegg to contig2kegg

		print join("\t", $gp_id, $kegg, $contig, $gstart, $gend);
		print "\n";

		$prev_gene = $gene;
		if (eof) {
			last LINE;
		}
	}

	close(KEGG);
}

sub analyze_file {
	my $this         = shift;
	my $label        = $this->label;
	my $level        = $this->level;
	my $FeatureCount = $this->feature_count;
	my ($kegg_map, $Coverage, $Length, $ResampledDataCount) = @_;

	my $WRITE_HASH = 0;

	my %Assigned;

	# Read the kegg_maps

	open(KEGG, "<$kegg_map") || die "Cannot open $kegg_map: $!";

	my $assigned=0;

	# Process the whole dataset, coverage needs to be used here, since this is the original data

	#my %KeggLength;

	# Process the kegg map file by calculating coverage of every kegg protein that predicted genes map to

	my $Gene2Kegg = {};	# array per kegg protein that tracks positions covered by sample
	my $KeggCoverage = {};
	my $KeggGeneInheritance = {};
	my $prev_gene = "dummy007";
	LINE:while (my ($gene, $kegg, $gstart, $gend, $kstart, $kend) = $this->parse_next_kegg_map_line(\*KEGG)) {
		last LINE unless $gene;
		if ($gene ne $prev_gene) { 

			# update running sums for all kegg proteins that this gene maps to
			# input here is a BEST-blast hit file. Multiple KEGG proteins per query is 
			# possible when the scores have a tie.
			# In those cases, we should avoid double-counting of two proteins from the same KO. 
			# So we divide by the number 
			# of KEGG genes each query maps to.

			my $keggs = scalar(keys %$Gene2Kegg);
			foreach my $k (keys %$Gene2Kegg) {
				my $overlap = 0;
				map {$overlap++ if ($Gene2Kegg->{$k}->[$_]);} (0..$#{$Gene2Kegg->{$k}});
				my $transferred = ($overlap*$Coverage->{$prev_gene}/$keggs);
				$KeggGeneInheritance->{$k}->{$prev_gene} += $transferred;
			}
			$Gene2Kegg = {};
		}
		map {$Gene2Kegg->{$kegg}->[$_] = 1} ($kstart..$kend);
		$prev_gene = $gene;
	}

	# process the final group after reaching end-of-file

	my $keggs = scalar(keys %$Gene2Kegg);
	foreach my $k (keys %$Gene2Kegg) {
		my $overlap = 0;
		map {$overlap++ if ($Gene2Kegg->{$k}->[$_]);} (0..$#{$Gene2Kegg->{$k}});
		my $transferred = ($overlap*$Coverage->{$prev_gene}/$keggs);
		$KeggGeneInheritance->{$k}->{$prev_gene} += $transferred;
	}

=begin DEBUG

write_two_column_hash("new_kegg_coverage_before.txt", \%KeggGeneInheritance);

=cut

	# normalize coverage by kegg protein length

	# Get protein lengths

	my $dbh = $this->smash->get_refgenomedb_handle;

	# scoping for $kegg_length_sth
	{ 
		my $kegg_length_sth = $dbh->prepare("SELECT protein, length FROM kegg_details");
		$kegg_length_sth->execute();
		my ($kegg, $length);
		$kegg_length_sth->bind_columns(\$kegg, \$length);
		while ($kegg_length_sth->fetch()) {
			if ($KeggGeneInheritance->{$kegg}) {
				foreach my $gene (keys %{$KeggGeneInheritance->{$kegg}}) {
					my $transferred = $KeggGeneInheritance->{$kegg}->{$gene} / $length;
					print "$gene\t$kegg\t$transferred\n";
					$KeggCoverage->{$kegg} += $transferred;
				}
			}
		}
	}

=begin DEBUG

write_two_column_hash("new_kegg_coverage_after.txt", \%KeggCoverage);

=cut

	if ($this->level ne "kegg") {
		my ($kegg, $ko);
		my ($count, $length);
		my $KOCoverage;

		# transfer coverage from kegg protein to ko

		my $sth = $dbh->prepare("SELECT protein, ko FROM protein2ko");
		$sth->execute();
		$sth->bind_columns(\$kegg, \$ko);
		while ($sth->fetch()) {
			if ($KeggCoverage->{$kegg}) {
				$KOCoverage->{$ko} += $KeggCoverage->{$kegg};
			}
		}

		# create feature vectors in KO

		$assigned = 0;
		while (my ($ko, $coverage) = each %$KOCoverage) {
			$FeatureCount->{$ko}->{$label} = $coverage;
			$assigned += $coverage;
		}

		# record feature vector if required

		if ($WRITE_HASH == 1) {
			my $genepred = $this->smash->genepred;
			open(KO, ">$genepred.sample_ko.txt");
			while (my ($ko, $coverage) = each %$KOCoverage) {
				print KO "$ko\t$label\t$coverage\n";
			}
			close(KO);
		}

		# track gene2function maps

	} else {

		# create feature vectors in KEGG proteins

		$assigned = 0;
		while (my ($kegg, $coverage) = each %$KeggCoverage) {
			$FeatureCount->{$kegg}->{$label} = $coverage;
			$assigned += $coverage;
		}

	}

	$this->smash->close_refgenomedb_handle();
	close(KEGG);
	return $assigned;
}

sub parse_next_kegg_map_line {
	my $this = shift;
	my $FH   = shift;
	my $line = $this->get_next_uncommented_line($FH);
	return 0 unless $line;
	my @w = split(/\s+/, $line);
	my ($gene, $kegg, $gstart, $gend, $kstart, $kend) = @w[0,1,17,18,20,21];
	return ($gene, $kegg, $gstart, $gend, $kstart, $kend);
}

sub convert_to_abundance_wrong {

	# convert from KO counts to KO abundance
	# I used to divide the KO abundance by the number of proteins in each KO, thinking EVERY hit above 60bits
	# was used. But apparently it is a best-blast-hit procedure, so I shouldnt do it. The only thing I should
	# do is to divide by the number of KEGG genes each predicted gene hits with the same score, to avoid 
	# double counting. So I removed the normalization of KO abundances by the KO protein_count. Mani, 18.11.2009

}

sub merge_feature_vectors {
	my $this          = shift;
	my $feature_count = shift;
	my $level         = shift;

	# merge if necessary

	if ($level eq "module" || $level eq "pathway") {
		return $this->merge_feature_vectors_internal($feature_count, "ko", $level);
	} elsif ($level eq "ko") {
		return $feature_count;
	}
}

sub merge_feature_vectors_internal {
	my $this          = shift;
	my $feature_count = shift;
	my $from          = shift;
	my $to            = shift;

	my $dbh           = $this->smash->get_refgenomedb_handle();

	my $merged_feature_count   = {};
	my $gene_physical_coverage = $this->gene_physical_coverage;

	my $WRITE_HASH = 0;

	# transfer coverage from ko to module

	if ($from eq "ko" && ($to eq "module" || $to eq "pathway") ||
	   ($from eq "protein" && $to eq "ko")) {
		my $function2gene = {};
		my $query = "SELECT $to FROM ${from}2$to WHERE $from=?";
		my $sth = $dbh->prepare($query);
		foreach my $k (keys %$feature_count) {
			$sth->execute($k);
			my @samples = keys %{$feature_count->{$k}};
			while (my ($function) = $sth->fetchrow_array()) {

				# propagate sample2ko mapping to this level
				foreach my $sample (@samples) {
					$merged_feature_count->{$function}->{$sample} += $feature_count->{$k}->{$sample};
				}
			}
		}

		if ($this->size_normalize == 1) {
			my %Count;
			my ($function, $count);
			$query = "SELECT $to, COUNT($from) FROM ${from}2$to GROUP BY $to";
			$sth = $dbh->prepare($query);
			$sth->execute();
			$sth->bind_columns(\$function, \$count);
			while ($sth->fetch()) {
				$Count{$function} = $count;
			}
			foreach my $function (keys %$merged_feature_count) {
				my $count = $Count{$function};
				foreach my $sample (keys %{$merged_feature_count->{$function}}) {
					$merged_feature_count->{$function}->{$sample} /= $count;
				}
				foreach my $gene (keys %{$function2gene->{$function}}) {
					$function2gene->{$function}->{$gene} /= $count;
				}
			}
		}
	}

	$this->smash->close_refgenomedb_handle();
	return $merged_feature_count;
}

1;

package Smash::Analyses::Comparative::Functional::eggNOG;
use strict;
use warnings;
use Smash::Utils::MatrixIO qw(read_R_matrix write_R_matrix transpose_matrix write_two_column_hash);

our @ISA = qw(Smash::Analyses::Comparative::Functional);

sub analyze_file {
	my $this         = shift;
	my $label        = $this->label;
	my $level        = $this->level;
	my $FeatureCount = $this->feature_count;
	my $genepred     = $this->smash->genepred;
	my ($eggnog_map, $Coverage, $Length, $ResampledDataCount) = @_;

	my $WRITE_HASH = 0;

	my %Assigned;

	# Read the kegg_maps

	open(STRING, "<$eggnog_map") || die "Cannot open $eggnog_map: $!";

	my $assigned=0;

	# Process the whole dataset, coverage needs to be used here, since this is the original data

	#my %KeggLength;


	# Process the kegg map file by calculating coverage of every kegg protein that predicted genes map to

	my $Gene2String = {};
	my %StringLength;
	my $StringCoverage = {};
	my %OverlapScore;
	my $prev_gene = "dummy007";
	LINE:while (my ($gene, $string, $sstart, $send, $cog, $length) = parse_next_eggNOG_map_line(\*STRING)) {

		$StringLength{$string} = $length; # record string protein's length. might do it multiple times, but it's ok!

		if (eof(STRING) || $gene ne $prev_gene) { # last line will match eof

			##################
			# each metagenome gene is mapped to one or more COG(s) via a string protein.
			# we keep track of the mappings to the COG and the protein by a 3-D array $Gene2String.
			# we then transfer the coverage from this gene to string proteins.
			# since multiple proteins can be used to map to the same COG in principle, 
			# we divide by the number of proteins this gene mapped to under this COG.
			##################

			foreach my $c (keys %$Gene2String) {                             # for every COG this gene maps to
				my $strings = scalar(keys %{$Gene2String->{$c}});        #   count number of string proteins under this COG that the gene maps to
				foreach my $s (keys %{$Gene2String->{$c}}) {             #   for every string protein under this COG
					my $overlap = 0;
					map {$overlap++ if ($Gene2String->{$c}->{$s}->[$_]);} (0..$#{$Gene2String->{$c}->{$s}});  # count the number of overlapping positions.
					$StringCoverage->{$cog}->{$s} += ($overlap*$Coverage->{$prev_gene}/$strings);             # if there are multiple string proteins, share coverage.
					                                                                                          # when we add them together for the COG later, this will make sense.
					$OverlapScore{$s}{$prev_gene} = $overlap;
				}
			}
			$Gene2String = {};
			if (eof(STRING)) {
				last LINE;
			}
		}
		map {$Gene2String->{$cog}->{$string}->[$_] = 1} ($sstart..$send);
		$prev_gene = $gene;
	}

#write_two_column_hash("kegg_coverage_before.txt", \%StringCoverage);

	# create feature vectors in COG

	########
	# each COG will probably have multiple proteins.
	# each COG gets coverage of each protein under it, normalized by the protein length.
	########

	$assigned = 0;
	foreach my $cog (keys %$StringCoverage) {
		my $sum = 0;
		foreach my $string (keys %{$StringCoverage->{$cog}}) {
			$sum += $StringCoverage->{$cog}->{$string}/$StringLength{$string};
		}
		$FeatureCount->{$cog}->{$label} = $sum;
		$assigned += $sum;
	}

	# record feature vector if required

	if ($WRITE_HASH == 1) {
		open(COG, ">$genepred.sample_cog.txt");
		close(COG);
	}

	close(STRING);
	return $assigned;
}

#####################
# eggNOG mapping output file looks as follows:
#
# gene, string_protein, COG, string_protein_start, string_protein_end, bit_score, string_protein_length
#
#####################

sub parse_next_eggNOG_map_line {
	my $fh = shift;
	my $line;

	return 0 if eof $fh;

	# get the next non-comment line

	do {
		$line = <$fh>;
	} while ($line =~ /^#/);

	return 0 unless $line;
	return 0 if ($line =~ /^#/);
	chomp($line);

	my ($gene, $string, $cog, $sstart, $send, $score, $length) = split(/\s+/, $line);
	return ($gene, $string, $sstart, $send, $cog, $length);
}

1;
