package Smash::Batch;

use strict;
use warnings;
use Fcntl;
use File::Temp;
use Smash::Global qw(:all);

sub new {
	my $class = shift;
	my $this  = bless {}, $class;
	$this->{TEMPFILE} = new File::Temp();
	$this->{OUTFILE}  = new File::Temp(DIR => ".", TEMPLATE => "batch.XXXXXXXX", SUFFIX => ".out", UNLINK => 0);
	printf STDERR "Smash::Batch instance: Output of this run will be reported in %s\n", $this->outfile;
	printf STDOUT "Smash::Batch instance: Output of this run will be reported in %s\n", $this->outfile;
	return $this;
}

sub tempfile {shift->{TEMPFILE}}
sub outfile  {shift->{OUTFILE}}

sub process_batch {
	my $this  = shift;
	my $flat  = shift;
	my @batch = split(/\n+/, $flat);
	print {$this->outfile} "<SmashBatch>\n";
	STEP: foreach my $step (@batch) {
		next STEP if $step =~ /^\s*#/ || $step =~ /^\s*$/;
		my ($mode, @rest) = split(/\s+/, $step);
		if ($mode eq "SET") {
			my $target  = shift @rest;
			$this->{VARIABLES}->{$target} = join(" ", @rest);
		} elsif ($mode eq "CAPTURE") {
			my $target  = shift @rest;
			my $script  = shift @rest;
			   $script  =~ s|^[\./]||;
			   $script  = "$SMASH_SCRIPT_LOCATION/$script";
			my $command = join(" ", $script, @rest);
			my $capture = $this->get_output($command);
			$this->{VARIABLES}->{$target} = $capture;
			die "$command failed to generate output\n" unless $capture;
		} elsif ($mode eq "SUCCEED") {
			my $script  = shift @rest;
			   $script  =~ s|^[\./]||;
			   $script  = "$SMASH_SCRIPT_LOCATION/$script";
			my $command = join(" ", $script, @rest);
			my $status  = $this->check_success($command);
			die "$command failed with status $status\n" unless $status==1;
		} elsif ($mode eq "SYSTEM") {
			my $command = join(" ", @rest);
			my $status  = $this->run_system_command($command);
			die "$command failed with status $status\n" unless $status==0;
		}
	}
	print {$this->outfile} "</SmashBatch>\n";
}

sub transform {
	my $this   = shift;
	my $string = shift;
	while (my ($key, $value) = each %{$this->{VARIABLES}}) {
		$string =~ s/<__${key}__>/$value/g;
	}
	return $string;
}

sub check_success {
	my $this    = shift;
	my $command = shift;
	my $output  = $this->run_smash_step($command);
	return 1 if $output and lc($output) eq "success";
	return 0;
}

sub get_output {
	my $this    = shift;
	my $command = shift;
	return $this->run_smash_step($command);
}

# No output/error capturing!

sub run_system_command {
	my $this    = shift;
	my $command = shift;
	   $command = $this->transform($command);
	print("$command\n");
	return system($command);
}

sub run_smash_step {
	my $this    = shift;
	my $command = shift;
	my $temp    = $this->tempfile;
	my $output;

	   $command = $this->transform($command);
	print("$command\n");
	system("$command 1>$temp 2>&1");
	seek $temp, SEEK_SET, 0;
	print {$this->outfile} "<SmashStep command='$command'>\n";
	LINE: while (<$temp>) {
		print {$this->outfile} $_;
		chomp();
		if (m|<output>(.*)</output>|) {
			$output = $1;
			last LINE;
		}
	}
	print {$this->outfile} "</SmashStep>\n";
	return $output;
}

1;
