#########################################################################################
#########################################################################################
## Generic cluster management
#########################################################################################
#########################################################################################

package Smash::Utils::Cluster;
use strict;
use warnings;
our @ISA = qw(Smash::Utils);

=head1 NAME

Smash::Utils::Cluster - General Cluster Management library

=head1 SYNOPSIS

	use Smash::Utils::Cluster;

	my $cluster = new Smash::Utils::Cluster::sigma();

=cut

############################################
# Constructor
############################################

sub new {
	my $class  = shift;
	my %params = @_;
	bless {
		%params
	}, $class;
}

use overload '""' => 'name';

sub name      { shift->{NAME}}
sub master    { shift->{MASTER}}
sub type      { shift->{TYPE}}
sub qsub      { shift->{QSUB}}
sub qstat     { shift->{QSTAT}}
sub max_nodes { shift->{MAX_NODES}}
sub parent    { shift->{PARENT}}
sub job_id    { shift->{JOB_ID}}
sub sub_task_id { shift->{SUB_TASK_ID}}
sub preamble  {}

1;

#########################################################################################
#########################################################################################
## PBS family
#########################################################################################
#########################################################################################

package Smash::Utils::Cluster::PBS;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster);

############################################
# Constructor
############################################

sub new {
	my $class      = shift;
	my %params     = @_;
	my $self       = $class->SUPER::new(@_);
	$self->{TYPE}  = "PBS";
	$self->{QSUB}  = "qsub"; 
	$self->{QSTAT} = "qstat -f";
	$self->{SUB_TASK_ID} = '$PBS_ARRAY_INDEX';
	$self->{JOB_ID} = '$PBS_JOBID';
	return $self
}

sub get_array_job_options {
	my $this       = shift;
	my $array_size = shift;
	if ($array_size == 1) {
		return "";
	}
	return "-J 1-$array_size";
}

sub get_cpu_options {
	my $this = shift;
	my $cpus = shift;
	return "-l select=1:ncpus=$cpus";
}

###########################################
# Check status of a job in the cluster
#
# returns:
#	0 - job is done
#	1 - job is still running
###########################################
sub get_job_status {
	my $this         = shift;
	my $job_id       = shift;
	my $qstat        = $this->qstat;
	my $output       = $this->execute_capture("$qstat $job_id");

	# Empty output usually means job doesnt exist
	if ($output eq "") { # job is finished!
		return 0;
	}

	# Output must contain at least a line with "job_state = "
	my @lines = split(/\n/, $output);
	my @state = grep {$_ =~ /job_state = /} @lines;

	# If "job_state = " was not found, may be server was unreachable
	if (scalar(@state) != 1) {
		warn "Error while checking status for job $job_id. Perhaps PBS master server was unreachable.";
	} else {
		my $real_state = $state[0];
		$real_state =~ s/job_state = //;
		$real_state =~ s/\s*$//;
		$real_state =~ s/^\s*//;
		if ($real_state eq "X") {
			return 0;
		}
	}
	return 1;
}

1;

#########################################################################################
#########################################################################################
## SGE family
#########################################################################################
#########################################################################################

package Smash::Utils::Cluster::SGE;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster);

sub new {
	my $class       = shift;
	my %params      = @_;
	my $self        = $class->SUPER::new(@_);
	$self->{TYPE}   = "SGE";
	$self->{QSUB}   = "qsub";
	$self->{QSTAT}  = "qstat";
	$self->{JOB_ID} = '$JOB_ID';
	$self->{SUB_TASK_ID} = '$SGE_TASK_ID';
	return $self;
}

sub get_array_job_options {
	my $this       = shift;
	my $array_size = shift;
	if ($array_size == 1) {
		return "";
	}
	return "-t 1-$array_size";
}

sub get_cpu_options {
	my $this = shift;
	die "SGE parallel environment is queue specific. Please implement get_cpu_options for $this\n";
}

###########################################
# Check status of a job in the cluster
#
# returns:
#	0 - job is done
#	1 - job is still running
###########################################
sub get_job_status {
	my $this         = shift;
	my $job_id       = shift;
	my $qstat        = $this->qstat;
	my $output       = $this->execute_capture("$qstat $job_id");
	if ($output eq "") { # job is finished!
		return 0;
	}
	my @lines = split(/\n/, $output);
	my @state = grep {$_ =~ /job_state = /} @lines;
	if (scalar(@state) > 1) {
		die "Error while checking status for job $job_id";
	}
	my $real_state = $state[0];
	$real_state =~ s/job_state = //;
	$real_state =~ s/\s*$//;
	$real_state =~ s/^\s*//;
	if ($real_state eq "X") {
		return 0;
	}
	return 1;
}

1;

#############################################################
#############################################################
##             EMBL site specific clusters                 ##
#############################################################
#############################################################

#############################################################
##                  PBS-master                             ##
#############################################################

package Smash::Utils::Cluster::EMBL;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster::PBS);

sub new {
	my $class  = shift;
	my $self   = $class->SUPER::new(@_);
	$self->{QSUB}   = "/usr/pbs/bin/qsub";
	$self->{QSTAT}  = "/usr/pbs/bin/qstat -f";
	$self->{MASTER} = "sub-master.embl.de";
	return $self;
}

sub qsub {
	my $this = shift;
	if (!$this->name) {
		die "EMBL cluster needs a name of the cluster!";
	}
	return $this->{QSUB}." -q ".$this->name;
}

sub preamble {
	return "source /usr/local/bin/pbs.sh;";
}

1;

#############################################################
##                  clusterng                              ##
#############################################################

package Smash::Utils::Cluster::clusterng;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster::EMBL);

sub new {
	my $class  = shift;
	my $self   = $class->SUPER::new(@_);
	if (!defined($self->{MAX_NODES})) {
		$self->{MAX_NODES} = 32;
	}
	$self->{NAME} = "clusterng";
	return $self;
}
1;

#############################################################
##                  parallel                               ##
#############################################################

package Smash::Utils::Cluster::parallel;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster::EMBL);

sub new {
	my $class  = shift;
	my $self   = $class->SUPER::new(@_);
	if (!defined($self->{MAX_NODES})) {
		$self->{MAX_NODES} = 32;
	}
	$self->{NAME} = "parallel";
	return $self;
}
1;

#############################################################
##                  mpi                                    ##
## from clnode181 to clnode196 with 8*16=128 cpus and 16GB ##
#############################################################

package Smash::Utils::Cluster::mpi;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster::EMBL);

sub new {
	my $class  = shift;
	my $self   = $class->SUPER::new(@_);
	if (!defined($self->{MAX_NODES})) {
		$self->{MAX_NODES} = 128;
	}
	$self->{NAME} = "mpi";
	return $self;
}
1;

#############################################################
##                  bork                                    ##
## from clnode239 to clnode294 with 8*56=448 cpus and 16GB ##
#############################################################

package Smash::Utils::Cluster::bork;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster::EMBL);

sub new {
	my $class  = shift;
	my $self   = $class->SUPER::new(@_);
	if (!defined($self->{MAX_NODES})) {
		$self->{MAX_NODES} = 448;
	}
	$self->{NAME} = "bork";
	return $self;
}
1;

#############################################################
##                  otto                                   ##
#############################################################

package Smash::Utils::Cluster::otto;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster::SGE);

sub new {
	my $class  = shift;
	my $self   = $class->SUPER::new(@_);
	if (!defined($self->{MAX_NODES})) {
		$self->{MAX_NODES} = 30;
	}
	$self->{MASTER} = "otto";
	$self->{QSUB}  = "/net/c1/local/sge/bin/lx24-amd64/qsub";
	$self->{QSTAT} = "/net/c1/local/sge/bin/lx24-amd64/qstat -j";
	return $self;
}

sub preamble {
	return "source /etc/profile.d/sge_settings.sh;";
}

sub get_cpu_options {
	my $this = shift;
	my $cpus = shift;
	return "-pe smp $cpus";
}

1;

#############################################################
##                  sigma                                  ##
#############################################################

package Smash::Utils::Cluster::sigma;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster::SGE);

sub new {
	my $class  = shift;
	my $self   = $class->SUPER::new(@_);
	if (!defined($self->{MAX_NODES})) {
		$self->{MAX_NODES} = 50;
	}
	$self->{MASTER} = "sigma";
	$self->{QSUB}  = "/usr/sgeadmin/bin/lx24-amd64/qsub";
	$self->{QSTAT} = "/usr/sgeadmin/bin/lx24-amd64/qstat -j";
	return $self;
}

sub preamble {
	return "source /etc/profile.d/settings.sh;";
}

sub get_cpu_options {
	my $this = shift;
	my $cpus = shift;
	return "-pe smp $cpus";
}

1;

#############################################################
##                  epsilon                                ##
#############################################################

package Smash::Utils::Cluster::epsilon;
use strict;
use warnings;
our @ISA = qw(Smash::Utils::Cluster::SGE);

sub new {
	my $class  = shift;
	my $self   = $class->SUPER::new(@_);
	if (!defined($self->{MAX_NODES})) {
		$self->{MAX_NODES} = 80;
	}
	$self->{MASTER} = "epsilon";
	$self->{QSUB}  = "/opt/gridengine/bin/lx26-amd64/qsub";
	$self->{QSTAT} = "/opt/gridengine/bin/lx26-amd64/qstat -j";
	return $self;
}

sub preamble {
	return "source /etc/profile.d/sge-binaries.sh;";
}

sub get_cpu_options {
	my $this = shift;
	my $cpus = shift;
	return "-pe smp $cpus";
}

1;
