package Smash::Utils::ClusterBLAST;
use strict;
use warnings;
use base("Smash::Utils");

sub run_blast_cluster {
	use File::Basename;
	use File::Temp;
	use File::Spec;
	use File::Glob;

	my $this          = shift;
	my %params        = @_;

	# From arguments
	my $blast         = $params{'BLAST'}        || die "BLAST undefined in ClusterBLAST";
	my $database      = $params{'BLASTDB'}      || die "BLASTDB undefined in ClusterBLAST";
	my $query         = $params{'QUERY'}        || die "QUERY undefined in ClusterBLAST";
			    (-f $query)             || die "Query file $query does not exist";
	my $name          = $params{'LABEL'}        || die "LABEL undefined in ClusterBLAST";
	my $output_dir    = $params{'OUTDIR'}       || die "OUTDIR undefined in ClusterBLAST";
			    (-d $output_dir)        || die "Directory $output_dir does not exist";
	my $pieces        = $params{'PIECES'}       || die "PIECES undefined in ClusterBLAST";

	my $E_threshold   = $params{'E'};
	                    defined($E_threshold)   || ($E_threshold = 0.1);
	my $bit_threshold = $params{'BITS'};
	                    defined($bit_threshold) || ($bit_threshold = 40);
	my $tabular       = $params{'TABULAR'};
	                    defined($tabular)       || ($tabular = 1);

	my $blast_params  = $params{'PARAMETERS'}   || "";
	my $cpus          = $params{'CPUS'}         || 1;
	my $start_only    = $params{'STARTONLY'}    || 0;
	my $max_subjects  = $params{'MAXSUBJECTS'}  || 10000;
	my $node_location = $params{'NODETEMP'}     || "/tmp/".(getpwuid($<)||"smash")."/in";


	# Locally decided
	my (undef, $script_location) = fileparse($0);
	my $cleanup_flag  = 0; # Just start the jobs and quit. Can combine later. Leave the temp folder alone
	my $db_name       = basename($database); # in case the database is somedir/somedb.x*
	my $query_name    = fileparse($query, qw(.fasta .fa .fna));
	my @files;
	my $file_count;
	my @commands = ();
	my $MAX_CONCURRENT_NFS_COPIES = 10;
	my $nfs_lock_counter = int(rand($MAX_CONCURRENT_NFS_COPIES)); # 0 <= $nfs_lock_counter < $MAX_CONCURRENT_NFS_COPIES

	my $shell_script = "$output_dir/$name.$blast.run.sh";
	open(SCRIPT, ">$shell_script") || die "Cannot open $shell_script: $!";

	####
	# Prolog
	####

###############################################################################################################################
#                                                                                                                             #
#                                                                                                                             #
		print SCRIPT<<EOS;
#!/bin/bash
piece=\$1;
nfs_lock=\$2;
sleep_time=\$3;
cpus=\$4;
if [ -z "\$piece" -o -z "\$nfs_lock" -o -z "\$sleep_time" ]; then
	echo "Usage: \$0 piece locknum sleep-time [cpus]" 1>&2;
	exit 2;
fi;
if [ -z "\$cpus" ]; then
	cpus=4;
fi;

sync_file() {
	local source=\$1;
	local local_copy=\$2;
	local lock=\$3
	if [ -z "\$source" -o -z "\$local_copy" ]; then
		echo "Usage: sync_file source copy" 1>&2;
		exit 2;
	fi;
	lockfile -r -1 -l 3600 \$source.lock.\$lock;
	lockfile -r -1 -l 3600 \$local_copy.lock; 
	if [ ! -f \"\$local_copy\" ]; then                               # file missing
		cp \$source \$local_copy; 
		cat \$local_copy | md5sum > \$local_copy.md5sum;
	fi;
	if [ ! -f \"\$local_copy.md5sum\" ]; then                        # md5sum missing
		cat \$local_copy | md5sum > \$local_copy.md5sum;
	fi;
	if [ -f \"\$local_copy\" ]; then                                 # file exists - either just copied, or existed before
		cat \$source | md5sum > \$local_copy.tmpsum;
		if ! diff -q \$local_copy.tmpsum \$local_copy.md5sum; then  # file is there, but md5sum differs
			cp \$source \$local_copy;
			cat \$local_copy | md5sum > \$local_copy.md5sum;
		fi;
		rm \$local_copy.tmpsum;
	fi;
	rm -f \$local_copy.lock;
	rm -f \$source.lock.\$lock;
}

sleep \$sleep_time;
mkdir -p $node_location;
EOS

#                                                                                                                             #
#                                                                                                                             #
###############################################################################################################################


	####
	# Sync database files
	####

	my @db_files = <$database.x??>;
	foreach my $source (@db_files) {
		my ($base, undef, $suffix) = fileparse($source, qr/\.[^.]*/);
		my $local_copy = "$node_location/$base$suffix";
###############################################################################################################################
#                                                                                                                             #
#                                                                                                                             #
		print SCRIPT "sync_file $source $local_copy \$nfs_lock;\n";
#                                                                                                                             #
#                                                                                                                             #
###############################################################################################################################
	}


	# Split query file into smaller files for cluster
	if ($pieces > 1) {
		my $split_dir     = File::Temp::tempdir("splitXXXXXXXX", DIR => $output_dir, CLEANUP => $cleanup_flag);
		my $split_suffix  = "fa";
		$this->execute("faSplit sequence $query $pieces $split_dir/") == 0 || die "Cannot run faSplit successfully";
		opendir(DIR, "$split_dir") || die "Cannot open $split_dir: $!";
		@files = map { "$split_dir/$_" } grep { -f "$split_dir/$_" } readdir(DIR);
		close(DIR);
		$file_count = scalar(@files);

		my ($split_subdir) = fileparse($split_dir);
		my $local_out_dir  = "/tmp/".(getpwuid($<)||"smash")."/out/$split_subdir";
####
# Final blast command
####

###############################################################################################################################
#                                                                                                                             #
#                                                                                                                             #
		print SCRIPT<<EOS;

target=@{[File::Spec->rel2abs($split_dir)]};
source=$local_out_dir;
mkdir -p \$source;
$blast $node_location/$db_name \$target/\$piece.$split_suffix -E=$E_threshold -cpus=\$cpus -B=$max_subjects -sort_by_highscore -warnings -notes -mformat=2 $blast_params > \$source/\$piece.$blast;
cp \$source/\$piece.$blast \$target/\$piece.$blast;
perl $script_location/filterBlastReport.pl --bits=$bit_threshold --sbjct=$max_subjects --check_complete --fasta=\$target/\$piece.$split_suffix \$source/\$piece.$blast > \$source/\$piece.$blast.filtered;
cp \$source/\$piece.$blast.filtered \$target/\$piece.$blast.filtered;
EOS
#                                                                                                                             #
#                                                                                                                             #
###############################################################################################################################


		####
		# JOBFILE has one command per line to run, preferably by a batch queue submission system.
		# It invokes the shell script generated earlier with each file.
		# PARTINFO contains information that can be used to combine these pieces together into a single blast report.
		####

		####
		# Figure out file names
		####

		my $queue_file = "$output_dir/$name.$blast.queue";
		my $part_file  = "$output_dir/$name.$blast.partinfo";
		open(JOBFILE, ">$queue_file") || die "Cannot open $queue_file: $!";
		open(PARTINFO, ">$part_file") || die "Cannot open $part_file: $!";

		####
		# Generate BLAST commands and write part info
		####

		print PARTINFO "$output_dir/$name.$blast\n";
		foreach my $input_file (@files) {
			my ($filename)  = fileparse($input_file, (".$split_suffix"));
			my $random_wait_seconds = int(rand(60));
			print JOBFILE "bash $shell_script $filename $nfs_lock_counter $random_wait_seconds $cpus\n";
			$nfs_lock_counter++;
			$nfs_lock_counter %= $MAX_CONCURRENT_NFS_COPIES;
			print PARTINFO "$input_file	$split_dir/$filename.$blast\n";
		}

		close(PARTINFO);
		close(JOBFILE);

		####
		# Summarize
		####
		print <<EOS;
PROGRESS
--------
Split $query into $file_count files in $split_dir
Wrote partition info into $output_dir/$name.partinfo
Wrote commands to $queue_file

SUCCESS

NEXT:
-----
	Submit $queue_file to the cluster, e.g., 
	enqueue.pl -N $name -mem 4000 --args="-l select=1:ncpus=$cpus" -q bork $queue_file

FINAL:
------
	Run the following command to merge pieces together:
		$script_location/combineBlastPieces.pl --bits=$bit_threshold $part_file
	or do it manually:
		cat $split_dir/*.$blast.filtered > $name.$blast
		cat $split_dir/*.$blast > $name.$blast.full
IN CASE OF ERROR:
-----------------
	If any individual piece needs to be reBLASTed, run the following command:
		bash $shell_script <piece> $nfs_lock_counter $cpus
EOS

	} else {
		my $local_out_dir  = "/tmp/".(getpwuid($<)||"smash")."/out";
		print SCRIPT<<EOS;
target=$output_dir;
source=$local_out_dir;
mkdir -p \$source;
$blast $node_location/$db_name $query -E=$E_threshold -cpus=$cpus -B=$max_subjects -sort_by_highscore -warnings -notes -mformat=@{[($tabular==1)?2:0]} $blast_params > \$source/$name.$blast;
cp \$source/$name.$blast \$target/$name.$blast;
EOS
		####
		# Summarize
		####
		print "SUCCESS:\n";
		print "--------\n";
		print "\tWrote commands to $shell_script\n";
		print "NEXT:\n";
		print "-----\n";
		print "\tRun it as follows:\n";
		print "\t\tbash $shell_script 00 $nfs_lock_counter $cpus\n";
	}
	close(SCRIPT);
}

sub file_sync_prologue {
	use File::Basename;
	my ($source, $local_copy) = @_;
	my (undef, $src_directory) = fileparse($source);
my $prologue = <<EOS;
	lockfile -r -1 -l 3600 $source.lock.\$nfs_lock;
	lockfile -r -1 -l 3600 $local_copy.lock; 
	if [ ! -f \"$local_copy\" ]; then                               # file missing
		cp $source $local_copy; 
		cat $local_copy | md5sum > $local_copy.md5sum;
	fi;
	if [ ! -f \"$local_copy.md5sum\" ]; then                        # md5sum missing
		cat $local_copy | md5sum > $local_copy.md5sum;
	fi;
	if [ -f \"$local_copy\" ]; then                                 # file exists - either just copied, or existed before
		cat $source | md5sum > $local_copy.tmpsum;
		if ! diff -q $local_copy.tmpsum $local_copy.md5sum; then  # file is there, but md5sum differs
			cp $source $local_copy; 
			cat $local_copy | md5sum > $local_copy.md5sum;
		fi;
		rm $local_copy.tmpsum;
	fi; 
	rm -f $local_copy.lock;
	rm -f $source.lock.\$nfs_lock;
EOS
	return $prologue;
}

sub combine_partial_blasts {
	my $part_info_file= shift;
	my $bit_threshold = shift;
	my $max_subjects  = shift || 10000;
	my @fasta_files   = ();
	my @blast_files  = ();
	my $output_file;
	my $line;

	# Read in the partition info
	open(PARTINFO, "<$part_info_file") || die "Cannot open $part_info_file: $!";
	if (defined($line = <PARTINFO>)) {
		chomp($line);
		$output_file = $line;
	}
	my @missing = ();
	while(<PARTINFO>) {
		chomp();
		my @words = split(/\t/);
		push(@fasta_files, $words[0]);
		push(@blast_files, $words[1]);
		if (! -f $words[1]) {
			warn "$words[1] missing\n";
			push(@missing, $words[1]);
		}
	}
	close(PARTINFO);
	if (@missing > 0) {
		die "Missing BLAST files\n";
	}

	# Sorted combined output file
	my $SORTED;
	open($SORTED, ">$output_file") || die "Cannot open $output_file: $!";
	for (my $i=0; $i<= $#blast_files; $i++) {
		# Read in each blast output file
		my $fasta_file = $fasta_files[$i];
		my $blast_file = $blast_files[$i];
		print "Processing $blast_file\n";
		check_complete_blast($blast_file, $fasta_file);
		write_sorted_blast_report("OUT_FH" => $SORTED, "BLAST_FILE" => $blast_file, "BITS" => $bit_threshold, "MAXSUBJECTS" => $max_subjects);
	}
	close($SORTED);
}

# Using MAXSUBJECTS will give all the hits that score as much as the top MAXSUBJECTS'th subject. 
# So it is safe to say MAXSUBJECTS => 1, since it will get all the top hits if they score exactly the same.
# Setting MAXSUBJECTS is important to trigger all the other filters. Since there is a lot of overhead, I have used
# MAXSUBJECTS as an indicator that any filtering will be done (except bitscore filter which we do a lot). So, if you want
# length, fraction, percent identity etc to be filtered on, pass MAXSUBJECTS with a very high value if necessary. Not setting
# MAXSUBJECTS will only filter based on bitscore.

sub write_sorted_blast_report {

	my %params = @_;

	my $bit_threshold;
	my $exp_threshold;
	my $max_subjects;
	my $sim_threshold;
	my $pid_threshold;
	my $min_query_aligned_bases;
	my $min_query_aligned_fraction;
	my $query_length_file;
	my $min_subject_aligned_bases;
	my $min_subject_aligned_fraction;
	my $subject_length_file;
	my $prev_query = "Manimozhiyan Arumugam";

	my @group          = ();
	my %QueryLengths   = ();
	my %SubjectLengths = ();

	# Mandatory

	my $blast_file                = $params{'BLAST_FILE'}  || die "BLAST undefined in ClusterBLAST";
	my $out_fh                    = $params{'OUT_FH'}      || die "Output file handler required";

	# Optional

	$bit_threshold                = $params{'BITS'}        || 40;
	$exp_threshold                = $params{'EXP'}         || 10;
	$sim_threshold                = $params{'SIM'}         || 10;
	$pid_threshold                = $params{'IDENTITY'}    || 10;
	$max_subjects                 = $params{'MAXSUBJECTS'};
	$min_query_aligned_bases      = $params{'MINQLEN'}     || 10;
	$min_query_aligned_fraction   = $params{'MINQFRAC'}    || 0.0;
	$query_length_file            = $params{'QLENFILE'};
	$min_subject_aligned_bases    = $params{'MINSLEN'}     || 10;
	$min_subject_aligned_fraction = $params{'MINSFRAC'}    || 0.0;
	$subject_length_file          = $params{'SLENFILE'};

	my $BLAST;
	if (ref $blast_file =~ /GLOB/) {    # argument is already a glob
		$BLAST = $blast_file;
	} elsif ($blast_file eq "-") {      # argument is "-", STDIN
		$BLAST = \*STDIN;
	} elsif ($blast_file =~ /\.gz$/) {  # argument is a gzipped file
		open($BLAST, "gunzip -c $blast_file |") || die "Cannot open gunzip pipe for $blast_file: $!";
	} else {                            # agument is a regular file
		open($BLAST, "<$blast_file") || die "Cannot open $blast_file: $!";
	}

	if (defined($query_length_file)) {
		open(FILE, "<$query_length_file") || die "Cannot open $query_length_file: $!";
		while(my $line=<FILE>) {
			chomp($line);
			my ($def, $len) = split(/\t/, $line);
			$QueryLengths{$def} = $len;
		}
		close(FILE);
	}

	if (defined($subject_length_file)) {
		open(FILE, "<$subject_length_file") || die "Cannot open $subject_length_file: $!";
		while(my $line=<FILE>) {
			chomp($line);
			my ($def, $len) = split(/\t/, $line);
			$SubjectLengths{$def} = $len;
		}
		close(FILE);
	}

	if (defined($max_subjects)) {
		LINE: while(my $line=<$BLAST>) {
			chomp($line);
			my @words = split(/\t/, $line);
			my ($query, $subject, $E, $bits, $score, $pcid, $pcpos, $qstart, $qend, undef, $sstart, $send) = (@words[0..2], @words[4..5], @words[10..11], @words[17..21]);
			#if ($query eq "query") {next LINE;}

			if ($bits < $bit_threshold || $E > $exp_threshold || $pcpos < $sim_threshold || $pcid < $pid_threshold) { next LINE;}

			
			{
				my $aligned_query_length = abs($qend - $qstart) + 1;
				if ($aligned_query_length < $min_query_aligned_bases) { next LINE;}
				if ($QueryLengths{$query} && $aligned_query_length / $QueryLengths{$query} < $min_query_aligned_fraction && $SubjectLengths{$subject} && ($SubjectLengths{$subject}-$send) > 10) { next LINE;}

				my $aligned_subject_length = abs($send - $sstart) + 1;
				if ($aligned_subject_length < $min_subject_aligned_bases) { next LINE;}
				if ($SubjectLengths{$subject} && $aligned_subject_length / $SubjectLengths{$subject} < $min_subject_aligned_fraction) { next LINE;}
			}

			if (eof || $query ne $prev_query) {

				# process the group of subjects for this query

				# if at end of file, add the line just read before you process the group of subjects

				push(@group, $line) if eof;

				# Sort the lines based on their scores

				my @scores = map {(split(/\t/, $_))[5]} @group;
				@group = @group[ sort {$scores[$b] <=> $scores[$a]} 0..$#group ]; # sort the lines based on the score of those lines
				@scores = (); # end context

				# Dont mess with the sort and map lines unless you know what you are doing
				# Get $max_subjects subjects in the sorted list

				# For each subject, we need to follow its max score in any HSP with this query
				
				my %MaxScore = ();
				@scores = ();
				for (@group) {
					my @words = split(/\t/);
					my $sbjct = $words[1];
					my $score = $words[5];
					if (!defined($MaxScore{$sbjct}) || ($MaxScore{$sbjct} < $score)) {
						$MaxScore{$sbjct} = $score;
						push(@scores, $score);
					}
				}

				# We now get a threshold that is the score of the $max_subjects'th score

				@scores = sort {$b <=> $a} @scores;
				my $threshold;
				if ($max_subjects-1 <= $#scores) {
					$threshold = $scores[$max_subjects-1];
				} else {
					$threshold = $scores[$#scores];
				}

				# Now we have: $MaxScore{"someone"} defined as max score for "someone"

				# find the subjects that have at least one hsp higher than threshold
				# then keep all their hsps
				# we dont care about the rest
				map {delete $MaxScore{$_} if $MaxScore{$_} < $threshold;} keys %MaxScore;
				@group = grep {defined($MaxScore{(split(/\t/))[1]})} @group;

				# Get all the subjects, since we need them to sort
				my @sbjcts = map {(split(/\t/, $_))[1]} @group;
				@group = @group[ sort {$MaxScore{$sbjcts[$b]} <=> $MaxScore{$sbjcts[$a]} || $sbjcts[$b] cmp $sbjcts[$a]} 0..$#group ]; # sort the lines based on the score of those lines and group subjects together as well

				# Print sorted lines
				for (@group) {
					print $out_fh "$_\n";
				}
				%MaxScore = ();

				@group = ();
			}
			push(@group, $line) if !eof; # if at the end of file, we have processed the last line already!
			$prev_query = $query;
		}

	} else {
		LINE: while(my $line=<$BLAST>) {
			chomp($line);
			my @words = split(/\t/, $line);
			my ($query, $subject, $E, $bits, $score, $pcid, $pcpos, $qstart, $qend, undef, $sstart, $send) = (@words[0..2], @words[4..5], @words[10..11], @words[17..21]);
			#if ($query eq "query") {next LINE;}

			if ($bits < $bit_threshold || $E > $exp_threshold || $pcpos < $sim_threshold || $pcid < $pid_threshold) { next LINE;}

			
			{
				my $aligned_query_length = abs($qend - $qstart) + 1;
				if ($aligned_query_length < $min_query_aligned_bases) { next LINE;}
				if ($QueryLengths{$query} && $aligned_query_length / $QueryLengths{$query} < $min_query_aligned_fraction && $SubjectLengths{$subject} && ($SubjectLengths{$subject}-$send) > 10) { next LINE;}

				my $aligned_subject_length = abs($send - $sstart) + 1;
				if ($aligned_subject_length < $min_subject_aligned_bases) { next LINE;}
				if ($SubjectLengths{$subject} && $aligned_subject_length / $SubjectLengths{$subject} < $min_subject_aligned_fraction) { next LINE;}
			}


			print $out_fh "$line\n";
		}
	}
	close($BLAST);
	my @scores = ();
	for (@group) {
		push (@scores, (split(/\t/, $_))[5]);
	}
	@group = @group[ sort {$scores[$b] <=> $scores[$a]} 0..$#group ];
	for (@group) {
		print $out_fh "$_\n";
	}
}

# Using MAXSUBJECTS will give all the hits that score as much as the top MAXSUBJECTS'th subject. 
# So it is safe to say MAXSUBJECTS => 1, since it will get all the top hits if they score exactly the same.
# Setting MAXSUBJECTS is important to trigger all the other filters. Since there is a lot of overhead, I have used
# MAXSUBJECTS as an indicator that any filtering will be done (except bitscore filter which we do a lot). So, if you want
# length, fraction, percent identity etc to be filtered on, pass MAXSUBJECTS with a very high value if necessary. Not setting
# MAXSUBJECTS will only filter based on bitscore.

sub write_sorted_blast_report_solexa_special {

	my %params = @_;

	my $bit_threshold;
	my $exp_threshold;
	my $max_subjects;
	my $sim_threshold;
	my $min_query_aligned_bases;
	my $min_query_aligned_fraction;
	my $query_length_file;
	my $min_subject_aligned_bases;
	my $min_subject_aligned_fraction;
	my $subject_length_file;
	my $prev_query = "Manimozhiyan Arumugam";

	my @group          = ();
	my %QueryLengths   = ();
	my %SubjectLengths = ();

	# Mandatory

	my $blast_file                = $params{'BLAST_FILE'}  || die "BLAST undefined in ClusterBLAST";
	my $out_fh                    = $params{'OUT_FH'}      || die "Output file handler required";

	# Optional

	$bit_threshold                = $params{'BITS'}        || 40;
	$exp_threshold                = $params{'EXP'}         || 10;
	$sim_threshold                = $params{'SIM'}         || 10;
	$max_subjects                 = $params{'MAXSUBJECTS'};
	$min_query_aligned_bases      = $params{'MINQLEN'}     || 10;
	$min_query_aligned_fraction   = $params{'MINQFRAC'}    || 0.0;
	$query_length_file            = $params{'QLENFILE'};
	$min_subject_aligned_bases    = $params{'MINSLEN'}     || 10;
	$min_subject_aligned_fraction = $params{'MINSFRAC'}    || 0.0;
	$subject_length_file          = $params{'SLENFILE'};

	my $BLAST;
	if (ref $blast_file =~ /GLOB/) {    # argument is already a glob
		$BLAST = $blast_file;
	} elsif ($blast_file eq "-") {      # argument is "-", STDIN
		$BLAST = \*STDIN;
	} elsif ($blast_file =~ /\.gz$/) {  # argument is a gzipped file
		open($BLAST, "gunzip -c $blast_file |") || die "Cannot open gunzip pipe for $blast_file: $!";
	} else {                            # agument is a regular file
		open($BLAST, "<$blast_file") || die "Cannot open $blast_file: $!";
	}

	if (defined($query_length_file)) {
		open(FILE, "<$query_length_file") || die "Cannot open $query_length_file: $!";
		while(my $line=<FILE>) {
			chomp($line);
			my ($def, $len) = split(/\t/, $line);
			$QueryLengths{$def} = $len;
		}
		close(FILE);
	}

	if (defined($subject_length_file)) {
		open(FILE, "<$subject_length_file") || die "Cannot open $subject_length_file: $!";
		while(my $line=<FILE>) {
			chomp($line);
			my ($def, $len) = split(/\t/, $line);
			$SubjectLengths{$def} = $len;
		}
		close(FILE);
	}

	if (defined($max_subjects)) {
		LINE: while(my $line=<$BLAST>) {
			chomp($line);
			my @words = split(/\t/, $line);
			my ($query, $subject, $E, $bits, $score, $pcpos, $qstart, $qend, undef, $sstart, $send) = (@words[0..2], @words[4..5], $words[11], @words[17..21]);
			#if ($query eq "query") {next LINE;}

			if ($E > $exp_threshold || $pcpos < $sim_threshold) { next LINE;}

			
			{
				my $aligned_query_length = abs($qend - $qstart) + 1;
				if ($aligned_query_length < $min_query_aligned_bases) { next LINE;}
				if ($QueryLengths{$query} && $aligned_query_length / $QueryLengths{$query} < $min_query_aligned_fraction) { next LINE;}

				my $aligned_subject_length = abs($send - $sstart) + 1;
				if ($aligned_subject_length < $min_subject_aligned_bases) { next LINE;}
				if ($SubjectLengths{$subject} && $aligned_subject_length / $SubjectLengths{$subject} < $min_subject_aligned_fraction) { next LINE;}
			}

			if ($query ne $prev_query && @group != 0) {
				# Sort the lines based on their scores
				my @bits;

				#my @scores = map {(split(/\t/, $_))[4]} @group;
				#@group = @group[ sort {$scores[$b] <=> $scores[$a]} 0..$#group ]; # sort the lines based on the score of those lines
				#@scores = (); # end context

				# Dont mess with the sort and map lines unless you know what you are doing
				# Get $max_subjects subjects in the sorted list

				# For each subject, we need to follow its max score in any HSP with this query
				
				my %MaxScore = ();
				for (@group) {
					my @words = split(/\t/);
					my $sbjct = $words[1];
					my $bit = $words[4];
					if (!defined($MaxScore{$sbjct}) || ($MaxScore{$sbjct} < $bit)) {
						$MaxScore{$sbjct} = $bit;
					}
				}

				# We now get a threshold that is the score of the $max_subjects'th score

				@bits = sort {$b <=> $a} values %MaxScore;
				my $threshold = 0;
				if ($max_subjects-1 <= $#bits) {
					$threshold = $bits[$max_subjects-1];
				} else {
					$threshold = $bits[$#bits];
				}

				if ($threshold < $bit_threshold) {
					$threshold = $bit_threshold;
				}

				# Now we have: $MaxScore{"someone"} defined as max score for "someone"

				# find the subjects that have at least one hsp higher than threshold
				# then keep all their hsps
				# we dont care about the rest
				map {delete $MaxScore{$_} if $MaxScore{$_} < $threshold;} keys %MaxScore;
				@group = grep {defined($MaxScore{(split(/\t/))[1]})} @group;

				# Get all the subjects, since we need them to sort
				my @sbjcts = map {(split(/\t/, $_))[1]} @group;
				@group = @group[ sort {$MaxScore{$sbjcts[$b]} <=> $MaxScore{$sbjcts[$a]} || $sbjcts[$b] cmp $sbjcts[$a]} 0..$#group ]; # sort the lines based on the score of those lines and group subjects together as well

				# Print sorted lines
				for (@group) {
					print $out_fh "$_\n";
				}
				%MaxScore = ();

				@group = ();
			}
			push(@group, $line);
			$prev_query = $query;
		}

	} else {
		LINE: while(my $line=<$BLAST>) {
			chomp($line);
			my @words = split(/\t/, $line);
			my ($query, $subject, $E, $bits, $score, $pcpos, $qstart, $qend, undef, $sstart, $send) = (@words[0..2], @words[4..5], $words[11], @words[17..21]);
			#if ($query eq "query") {next LINE;}

			if ($bits < $bit_threshold) { next LINE;}

			print $out_fh "$line\n";
		}
	}
	close($BLAST);
	my @scores = ();
	for (@group) {
		push (@scores, (split(/\t/, $_))[5]);
	}
	@group = @group[ sort {$scores[$b] <=> $scores[$a]} 0..$#group ];
	for (@group) {
		print $out_fh "$_\n";
	}
}

# return -1 on failure
sub check_complete_blast {
	use FAlite;
	my ($blast_file, $fasta_file) = @_;
	my $expected_words = 0;
	my @words = ();
	my $query;

	if (! -f $blast_file) {
		warn "$blast_file not found.\n";
		return -1;
	} elsif ( -z $blast_file) {
		warn "$blast_file looks corrupt. It has 0 bytes.\n";
		return -1;
	}

	# Check the BLAST file itself for errors.
	open(BLAST, "<$blast_file") || die "Cannot open $blast_file to read: $!";
	my $line;
	if (defined($line = <BLAST>)) {
		chomp($line);
		@words = split(/\t/, $line);
		$expected_words = scalar(@words);
	}
	while (<BLAST>) {
		chomp();
		$line = $_;
	}
	close(BLAST);

	# Check number of words in the last line
	@words = split(/\t/, $line);
	if (scalar(@words) != $expected_words) {
		warn "$blast_file looks corrupt. Found ".scalar(@words)." words instead of $expected_words.\n";
		return -1;
	}
	$query = $words[0];

	# See how many queries have been finished
	open(FASTA, "<$fasta_file") || die "Cannot open $fasta_file: $!";
	my $fasta = new FAlite(\*FASTA);
	my $entries = 0;
	my $blast_count = 0;
	while (my $entry = $fasta->nextEntry) {
		$entries++;
		my $def = $entry->def;
		$def =~ s/^>//;
		$def =~ s/\s.*//;
		if ($def eq $query) {
			$blast_count = $entries;
		}
	}
	close(FASTA);
	if ($blast_count / $entries < 0.9) {
		warn "$blast_file looks incomplete. BLAST finished at query $blast_count of $entries.\n";
		return -1;
	}
	return 1;
}

1;
