#!/usr/bin/env perl

##################################################################
#   Smash - Simple metagenomic analysis shell                    #
#   Copyright (c) 2008, EMBL-Heidelberg, Germany                 #
#   All Rights Reserved                                          #
#   Send all comments to Mani - arumugam@embl.de                 #
##################################################################

use strict;
use warnings;
use Pod::Usage;
use Getopt::Long;
use File::Basename;
use Smash::Utils::MatrixIO qw(:all);

use Smash::Core;
use Smash::Global qw(:all);
use Smash::CommandLineParser qw(parse_options check_required_options print_options);
use Smash::Utils::Taxonomy qw(:all);
use Smash::Utils::iTOL;
use Smash::Analyses::BootStrap;
use Smash::Analyses::Comparative;

my $UNIFORM_PRIOR = 0.000001;
my $REPLICATES = 100;

##############
# Set up command line parsing
##############

my @allowed  = qw(tag=s analysis=s experiment=s mode=s list=s level=s normalize=s identity=n replicate=n genomesize! filterlow=f dbfile=s conf_threshold=f length_threshold=i tree=s local help); # arguments I expect
   @allowed  = (@allowed, "refdb=s{,}"); 
my @required = qw(tag analysis mode list);   # arguments I require

##############
# Parse command line options
##############

my $status;
my $missing;
my %options;

($status, %options) = parse_options(\@allowed);
if ($options{help}) {
	pod2usage(-exitstatus => 0, -verbose => 2);
}
if ($status != 1) {
	pod2usage(-message => "", -exitstatus => 2, -verbose => 1);
}
#print_options(%options);
($status, $missing) = check_required_options(\@required, %options);
if ($status != 1) {
	pod2usage(-message => "$SMASH_SCRIPT_NAME: Missing argument --$missing\n", -exitstatus => 2, -verbose => 1);
}
########################
# Command line arg vars
########################


my @opt_refdb = @{$options{refdb}};
my @opt_distance;

# defaults

# genome size normalize is on unless --nogenomesize is specified

$options{genomesize} = 1 unless defined($options{genomesize});
$options{tree}       = "ncbi" unless $options{tree};


########################
# Parse command line options
########################

####
# Check proper usage
####

if (index(":functional:refgenome:16S:evodistance:refcoverage:refbasecoverage:keggrescoverage:", ":$options{analysis}:") == -1) {
	pod2usage(-message => "$SMASH_SCRIPT_NAME: Invalid option --analysis=$options{analysis}", -exitstatus => 2, -verbose => 1);
}

if ($options{analysis} eq "refgenome" && $options{mode} eq "raw" && !defined($options{identity})) {
	pod2usage(-message => "$SMASH_SCRIPT_NAME: --analysis=refgenome --mode=raw needs --identity to be specified!", -exitstatus => 2, -verbose => 1);
}

############################################
# fix command line parameters
############################################

$options{experiment} = "metagenomic" unless $options{experiment};


############################################
# Common
############################################

my $COV_MIN = 0.05;
my $COV_MAX = 1.04;
my $COV_STEP = 0.05;
my @COVERAGES;
for (my $cov=$COV_MIN; $cov < $COV_MAX; $cov+=$COV_STEP) {
	push(@COVERAGES, $cov);
}

my $bootstrap_item;   # Item to bootstrap: reads, genes or proteins etc
my $bootstrap_label;  # Name of dir under $read_dir where bootstrap lists are kept
my $GLOBAL_TOP_VAL;   # Y axis range for itol/gnuplot graph
my $analyze;          # reference to the analysis function depending on $options{analysis}
my $init_analysis;
my $finish_analysis;
my $reporter;
my $analyzer;
my $dbh;
my $seq_sth;
my $refdb;            # Used to generate files in refgenome
my $PRINT_SCALE = 0;  # Add a dummy species to print scale

my $READ_PHYLO_MAP;   # File-handle for writing "read\ttax_id\n" information

my %Sequence2Tax; # Map reference sequences to tax_id

my %Labels;
my %Data; # for boxplots
my %NodesForTree;
my %DataCount = (); # Gene or read count
my %Assigned = ();
my @entities;

my $PROGRESS = \*STDERR;
select($PROGRESS); $| = 1; select(STDOUT);

my $sample                = "metagenome";
my $itol_prefix           = "$options{tag}.$options{analysis}.$options{mode}";
my $itol_main_config_file = "$itol_prefix.main.cfg";     # File for the iTOL config file
my $CONFIG;                                              # File-handle for the iTOL config file

# for smash paper and MetaHIT
my %Colors = (
		# 41 sanger samples
		"AM-A" => "#808284", "DA-A" => "#0000FF", "ES-A" => "#000000", "FR-A" => "#006E3A", 
		"IT-A" => "#00ADEF", "JP-A" => "#8E191B", "JP-I" => "#F26522", "AM-F" => "#8CC63F", 

		# sanger/454 comparison
		#"DA-AD-1" => "#8E191B", "DA-AD-3" => "#006E3A",

		# Smash paper figure
		"MN-SOIL" => "#000000", "EBPR" => "#006E3A", "WF-A" => "#00ADEF", "AMD" => "#F26522",

		# misc
		"assigned" => "#AA5B48", "unassigned" => "#B8D7EF",
		);


#################################################################
# Get the metagenome details
# @entities contains the list
# %Labels{$entity} contains the label of $entity
#################################################################

open(LIST, "<$options{list}") || die "Cannot open $options{list}: $!";
my @lines = <LIST>;
close(LIST);

# Get the labels for the samples/entities
# This could come from the input list file, then dont worry about the database

%Labels = ();
%DataCount = ();
@entities = ();
map {chomp($_); my ($entity, $count, $label) = split(/\t/); $Labels{$entity} = $label; $DataCount{$entity} = $count; push(@entities, $entity);} @lines;

# for local mode, data count must be passed from the file

if ($options{local} && scalar(grep {defined} values %DataCount) < 0) {
	die "List file $options{list} must contain datacount for --local\n";
}

# if labels werent defined, either get from database, or if local, assign id as label

if (scalar(grep {defined} values %Labels) == 0) { # labels are not given in input file
	if ($options{local}) {
		map {$Labels{$_} = $_} keys %Labels; # for genome, this will be undef, so we use the entity as label as well
	} else {
		my $smash = new Smash::Core();
		for (my $e = 0; $e < scalar(@entities); $e++) {
			my $entity = $entities[$e];
			$smash->init($entity);
			my $label = $smash->get_metagenome_label($smash->metagenome);
			$Labels{$entity} = $label || $entity; # for genome, this will be undef, so we use the entity as label as well
		}
		$smash->finish();
	}
}


@opt_distance = ("jsd"); #, "euclidean");

if ($options{mode} eq "cluster") {
	run_clustering([values %Labels], "neighbor", @opt_distance);
	success(0);
}

if ($options{mode} eq "bootstrap+cluster") {
	run_bootstrap_clustering([values %Labels], "neighbor", @opt_distance);
	success(0);
}


if ($options{analysis} eq "refgenome") {
	$bootstrap_label = "reads_refgenomes";
	$bootstrap_item  = "reads";
	$GLOBAL_TOP_VAL  = 23;
	if (!@opt_refdb) {
		@opt_refdb  = qw(ALL);
	}
} elsif ($options{analysis} eq "functional") {
	$bootstrap_label = "genes_functional";
	$bootstrap_item  = "genes";
	$GLOBAL_TOP_VAL  = 5;
	if (!@opt_refdb) {
		@opt_refdb  = qw(kegg);
	}
} elsif ($options{analysis} eq "16S") {
	$bootstrap_label = "reads_16S";
	$bootstrap_item  = "reads";
	$options{level}  = "genus" unless $options{level}; # default level is genus
	$GLOBAL_TOP_VAL  = 50;
	if (!@opt_refdb) {
		@opt_refdb  = qw(RDP);
	}
}

if ($options{analysis} eq "refgenome" || $options{analysis} eq "16S") {
	my $smash = new Smash::Core();
	$smash->init();
	#$NCBI_TAXONOMY_LOCAL_LOCATION  = $smash->get_conf_value("Taxonomy", "local_repository");
	#$NCBI_TAXONOMY_REMOTE_LOCATION = $smash->get_conf_value("Taxonomy", "remote_repository");
	$smash->finish();
}

# Init analysis type
init_reporter();

#################################################################
# Process each sample
#################################################################

REFDB: for (my $i=0; $i<@opt_refdb; $i++) {
	$refdb = $opt_refdb[$i];

	####################
	# set up analyzer
	####################

	my $class = get_analyzer($options{analysis});
	my $analyzer = $class->new(NAME => $options{tag}, TYPE => $options{analysis}, EXPERIMENT => $options{experiment}, NORMALIZE => $options{normalize}, LEVEL => $options{level}, IDENTITY => $options{identity}, REF_DB => $refdb, REPLICATE => $options{replicate}, IS_LOCAL => $options{local}, LOCAL_DB_FILE => $options{dbfile}, GENOMESIZE_NORMALIZE => $options{genomesize}, FILTERLOW => $options{filterlow}, CONFIDENCE_THRESHOLD => $options{conf_threshold}, READ_LENGTH_THRESHOLD => $options{length_threshold}, TREE => $options{tree});
	$analyzer->init();
	print $PROGRESS "Processing samples ";
	for (my $e = 0; $e < scalar(@entities); $e++) {

		####################
		# Convert from sample id to user-friendly labels
		# Attach this label/entity as the current sample
		####################

		my $entity = $entities[$e];
		my $label  = $Labels{$entity};
		if ($sample eq "metagenome") {
			$analyzer->attach_sample($entity, $label);
		} elsif ($sample eq "genome") {
			$analyzer->attach_genome($entity, $label);
		} else {
		}

		####################
		# run the analysis:
		####################

		if ($options{mode} eq "raw") {

			if ($sample eq "metagenome") {
				$Assigned{$label} = $analyzer->analyze_one_sample($bootstrap_label);
			} elsif ($sample eq "genome") {
				$Assigned{$label} = $analyzer->analyze_one_genome();
			}

			####################
			# We have used the datacount properly for analysis.
			# Now we might have to modify it for normalization purposes!
			####################

			if ($options{normalize} eq "hits") {
				$analyzer->{DATA_COUNT}->{$label} = $Assigned{$label};
			} else {
				my $this_data_count;
				if ($options{local}) {
					$this_data_count = $DataCount{$entity};
				} else {
					$this_data_count = $analyzer->get_data_count($entity);
				}
				$analyzer->{DATA_COUNT}->{$label} = $this_data_count;
				$analyzer->{FEATURE_COUNT}->{-1}->{$label} = $this_data_count - $Assigned{$label};
			}

			####################
			# Now adjust the datacount so that anything less than 100bp, which doesnt have a chance, is not included!
			####################

			if ($options{analysis} eq "refcoverage") {
				my $dbh;
				my $sth = $dbh->prepare_cached("SELECT COUNT(*) FROM readinfo r INNER JOIN library l USING (library_id) INNER JOIN sample s USING (sample_id) WHERE metagenome_id=? AND r.length >= ?");
				$sth->execute($entity, 100);
				my ($count) = $sth->fetchrow_array();
				$analyzer->data_count->{$label} = $count;
			}
		}
		print $PROGRESS Smash::Core->progress_bar($e+1);

	}
	print $PROGRESS " done\n";

	########################
	# if mode=raw, write the counts and move to the next refdb
	########################

	my $raw_file = "$options{tag}.feature_vector.rawcount.$options{analysis}.$refdb";
	if ($options{mode} eq "raw") {
		zero_fill_matrix($analyzer->feature_count);
		write_R_matrix($raw_file, $analyzer->feature_count);
		$analyzer->finish();
		next REFDB;
	}

	########################
	# read the rawcounts in
	########################

	if (! -f $raw_file) {
		die "Raw file $raw_file missing. Perhaps you need to run $SMASH_SCRIPT_NAME with '--mode=raw'?";
	}

	my $feature_count = read_R_matrix($raw_file);

	########################
	# filter the matrix and remove things not in the list file
	########################

	print $PROGRESS "Filtering samples ...";
	$feature_count = transpose_matrix($feature_count);
	{
		my @keys = keys %$feature_count;
		my %keep = map {$_ => 1} values %Labels;
		map {delete $feature_count->{$_} unless defined($keep{$_})} @keys;
	}
	$feature_count = transpose_matrix($feature_count);
	print $PROGRESS " done\n";

	$analyzer->{FEATURE_COUNT} = $feature_count;


	########################
	# For bootstrapping, we need the raw counts. So we do that before we convert things to abundance.
	########################

	if ($options{mode} eq "bootstrap+distmat") {
		$analyzer->bootstrap_elements();
		$analyzer->finish();
		next REFDB;
	}

	########################
	# Convert to abundance
	# Merge into higher levels, if necessary
	########################

	# normalize by genome size to convert to abundance

	if ($options{genomesize}) {
		$analyzer->convert_to_abundance($analyzer->feature_count); # now %FeatureCount is abundance
	}

	if ($analyzer->level) {
		$analyzer->{FEATURE_COUNT} = $analyzer->merge_feature_vectors($analyzer->feature_count, $analyzer->level); # now %FeatureCount is merged
	}
	zero_fill_matrix($analyzer->feature_count);

	my $file;

	########################
	# do the analysis
	########################

	if ($options{mode} eq "summary") {

		my $copy = copy_matrix($analyzer->feature_count);

		$copy = $analyzer->convert_taxid_to_name($copy); # get names in the matrix
		$copy = normalize_cols_by_sum($copy);    

		$file = sprintf("%s.feature_vector.summary.%s.%s.%s", $options{tag}, $options{analysis}, (($options{identity})?($options{identity}."p."):"").$options{level}, $refdb);
		write_R_matrix($file, $copy);
		
		# Special mode for Jeroen's over/under analysis
		# raw counts, genome size in Mb normalized, not sample size normalized
		# taxa by name

		$copy = $analyzer->convert_taxid_to_name($analyzer->feature_count);    # get names in the matrix
		my $factor = 1000000;
		if ($options{analysis} eq "16S") {
			$factor = 4;
		}
		scalar_multiply_matrix($copy, $factor);
		$file = "$options{tag}.$options{level}.for_jeroen.txt";
		write_R_matrix($file, $copy);

	} elsif ($options{mode} eq "distmat") {
		########################
		# For creating distance matrix, we DONT REALLY need the raw counts. 
		# But this is where I can create the raw counts for the non-raw levels (like module/pathway or genus/phylum).
		# So we do that before we convert things to abundance.
		########################

		# raw counts, genome size in bp normalized, not sample size normalized
		# taxa by tax_id

		$file = sprintf("%s.feature_vector.rawcount.%s.%s.%s", $options{tag}, $options{analysis}, (($options{identity})?($options{identity}."p."):"").$options{level}, $refdb);
		write_R_matrix($file, $analyzer->feature_count);

		if ($options{normalize} eq "hits") {
			delete $analyzer->{FEATURE_COUNT}->{-1};
		}
		$analyzer->{FEATURE_COUNT} = normalize_cols_by_sum($analyzer->feature_count);
		if ($analyzer->filterlow) {
			$analyzer->filter_low_occurrences($analyzer->filterlow, $analyzer->feature_count);
		}

		# raw counts, genome size in bp normalized, sample size normalized
		# filtered for low occurrences, if applicable
		# taxa by tax_id

		$file = sprintf("%s.feature_vector.normalized.%s.%s.%s", $options{tag}, $options{analysis}, (($options{identity})?($options{identity}."p."):"").$options{level}, $refdb);
		write_R_matrix($file, $analyzer->feature_count);

		$analyzer->make_distance_matrices($analyzer->feature_count);
	} elsif ($options{mode} eq "itol") {
		$analyzer->{FEATURE_COUNT} = normalize_cols_by_sum($analyzer->feature_count);
		if ($analyzer->filterlow) {
			$analyzer->filter_low_occurrences($analyzer->filterlow, $analyzer->feature_count);
		}

		my $fc = $analyzer->feature_count;
		my $new_fc = {};
		foreach my $rdp_tax_id (keys %$fc) {
			my $ncbi = bergey2ncbi($rdp_tax_id);
			$new_fc->{$ncbi} = $fc->{$rdp_tax_id} if $ncbi;
		} 
		$analyzer->{FEATURE_COUNT} = $new_fc;

		$reporter->add_dataset(NAME => $refdb, DATA => $analyzer->feature_count, GRAPH_TYPE => "barplot", SCALE => 100, KEEP_TOP_N => $options{top}); # convert to percent from fraction
		#$reporter->add_dataset(NAME => $refdb, DATA => $analyzer->feature_count, GRAPH_TYPE => "heatmap", SCALE => 100, KEEP_TOP_N => $options{top}); # convert to percent from fraction
	} elsif ($options{mode} eq "boxplot") {
		make_boxplot_files();
	} elsif ($options{mode} eq "refbasecoverage") {
		#compare_replicates_with_reference_base();
	} elsif ($options{mode} eq "refcoverage") {
		compare_replicates_with_reference(\%DataCount);
	} elsif ($options{mode} eq "qualdistmat") {
		make_qualitative_distance_matrices($analyzer->feature_count, $options{replicate});
	}
	$analyzer->finish();
}

finish_reporter();

success(0);

############################################
# Subroutines
############################################


sub success {
	print "<output>success</output>\n";
	exit(shift);
}

# Mani, 01 Apr 2010. No, not a joke!
# run_bootstrap()
# Old style bootstrapping. Will keep it for a while, but if we
# dont use it for a while, then should be deleted from here.

sub run_bootstrap {
	use Config;

	# Bootstrap

	Smash::Analyses::BootStrap::bootstrap_list(@_);

	# Create job script for replicate analysis

	my $perl = $Config{perlpath};
	my $replicate_job_file = "$options{tag}.replicates.queue";
	open(JOB, ">$replicate_job_file") || die "Cannot open $replicate_job_file";
	for (my $i=0; $i<$REPLICATES; $i++) {
		print JOB "$perl $0 --analysis=$options{analysis} --tag=$options{tag} --identity=$options{identity} --normalize=$options{normalize} --list=$options{list} --mode=distmat --replicate=$i --refdb ".join(" ", @opt_refdb);
		if (defined($options{level})) {
			print JOB " --level=$options{level}";
		}
		print JOB "\n";
	}
	close(JOB);
	print $PROGRESS "I have generated a script '$replicate_job_file' to process all the replicates.\n";
	print $PROGRESS "Run the script to generate all the matrices\n";
}

sub run_clustering {
	use File::Copy;

	my $labels  = shift;
	my $method  = shift || die "clustering needs method!";
	my @metrics = @_;
	die "clustering needs metric!" if @metrics < 1;

	my %keep = map {$_ => 1} @$labels; # samples to keep

	foreach my $metric (@metrics) {
		foreach my $refdb (@opt_refdb) {

			#####
			# Distance matrix
			#####

			open(DIST, ">infile") || die "Cannot open infile: $!";
			my $dist = read_phylip_matrix("$options{tag}.distance.$metric.$options{analysis}.$options{level}.$refdb");
			my @keys = keys %$dist;

			# delete unlisted in the row
			map {delete $dist->{$_} unless defined $keep{$_}} @keys;

			# delete unlisted in the column (row after transpose)
			$dist = transpose_matrix($dist);
			map {delete $dist->{$_} unless defined $keep{$_}} @keys;

			$dist = transpose_matrix($dist);
			write_phylip_matrix(\*DIST, $dist);
			close(DIST);

			#####
			# Run clustering method
			#####

			unlink qw(outfile outtree);
			my $seed = 1+2*int(rand(49));
			open(NEIGHBOR, "| $method");
			print NEIGHBOR<<EOF;
2
y

EOF
			close(NEIGHBOR);

			#####
			# clean up the tree
			#####

			my $tree = "";
			open(TREE, "<outtree");
			while (<TREE>) {
				chomp();
				$tree .= $_;
			}
			close(TREE);
			unlink "outtree";
			$tree =~ s/\.0//g;
			$tree =~ s/\):/\)#/g;
			$tree =~ s/:100//g;
			$tree =~ s/#//g;

			#####
			# Make the new clean tree
			#####

			open(TREE, ">$options{tag}.$options{analysis}.tree");
			print TREE "$tree\n";
			close(TREE);

			#####
			# Upload to iTOL
			#####

			my $treeName = sprintf("$options{analysis} clustering%s (%sDatabase:%s; Metric:%s; UniformPrior: %f; Method:%s;)", ($options{tag})?" $options{tag}":"", ($options{level})?"Level:$options{level}; ":"", $refdb, $metric, $UNIFORM_PRIOR, $method);

			my $smash = new Smash::Core();
			$smash->init();
			my %reporter_options = (
						    NAME => "$options{tag}.$options{analysis}",
						  LABELS => [map {$Labels{$_}} @entities],
					     DESCRIPTION => $treeName,
					     GRAPH_WIDTH => 25,
					NCBI_TAXONOMY_ID => 0,
					       UPLOAD_ID => $smash->get_conf_value("iTOL", "uploadID")
					);
			$smash->finish();


			$reporter = new Smash::Utils::iTOL(%reporter_options);
			if (-f "sample.cluster.itol.colors") {
				$reporter->{COLOR_FILE} = "sample.cluster.itol.colors";
			}

			$reporter->init();
			$reporter->upload();
			$reporter->download("svg");
		}
	}
	success(0);
}

sub run_bootstrap_clustering {
	use File::Copy;

	my $labels  = shift;
	my $method  = shift || die "clustering needs method!";
	my @metrics = @_;
	die "clustering needs metric!" if @metrics < 1;

	my %keep = map {$_ => 1} @$labels; # samples to keep

	foreach my $metric (@metrics) {
		foreach my $refdb (@opt_refdb) {

			#####
			# Distance matrix
			#####

			open(DIST, ">infile") || die "Cannot open infile: $!";
			for (my $i=0; $i<$REPLICATES; $i++) {
				my $dist = read_phylip_matrix("$options{tag}.distance.$metric.$options{analysis}.$options{level}.$refdb.$i");
				my @keys = keys %$dist;

				# delete unlisted in the row
				map {delete $dist->{$_} unless defined $keep{$_}} @keys;

				# delete unlisted in the column (row after transpose)
				$dist = transpose_matrix($dist);
				map {delete $dist->{$_} unless defined $keep{$_}} @keys;

				$dist = transpose_matrix($dist);
				write_phylip_matrix(\*DIST, $dist);
			}
			close(DIST);

			#####
			# Run clustering method
			#####

			unlink qw(outfile outtree);
			my $seed = 1+2*int(rand(49));
			open(NEIGHBOR, "| $method");
			print NEIGHBOR<<EOF;
m
$REPLICATES
$seed
2
y

EOF
			close(NEIGHBOR);
			move qw(outtree intree);

			#####
			# run consense
			#####

			unlink qw(outfile);
			open(CONSENSE, "| consense");
			print CONSENSE<<EOF;
2
y

EOF
			close(CONSENSE);

			#####
			# clean up the tree from consense
			#####

			my $tree = "";
			open(TREE, "<outtree");
			while (<TREE>) {
				chomp();
				$tree .= $_;
			}
			close(TREE);
			unlink "outtree";
			$tree =~ s/\.0//g;
			$tree =~ s/\):/\)#/g;
			$tree =~ s/:100//g;
			$tree =~ s/#//g;

			#####
			# Make the new clean tree
			#####

			open(TREE, ">$options{tag}.$options{analysis}.bootstrapcluster.tree");
			print TREE "$tree\n";
			close(TREE);

			#####
			# Upload to iTOL
			#####

			my $treeName = sprintf("$options{analysis} clustering%s (%sDatabase:%s; Metric:%s; UniformPrior: %f; Method:%s;)", ($options{tag})?" $options{tag}":"", ($options{level})?"Level:$options{level}; ":"", $refdb, $metric, $UNIFORM_PRIOR, $method);

			my $smash = new Smash::Core();
			$smash->init();
			my %reporter_options = (
						    NAME => "$options{tag}.$options{analysis}.bootstrapcluster",
						  LABELS => [map {$Labels{$_}} @entities],
					     DESCRIPTION => $treeName,
					     GRAPH_WIDTH => 25,
					NCBI_TAXONOMY_ID => 0,
					       UPLOAD_ID => $smash->get_conf_value("iTOL", "uploadID")
					);
			$smash->finish();


			$reporter = new Smash::Utils::iTOL(%reporter_options);
			if (-f "sample.cluster.itol.colors") {
				$reporter->{COLOR_FILE} = "sample.cluster.itol.colors";
			}

			$reporter->init();
			$reporter->upload();
			$reporter->download("svg");
		}
	}
	success(0);
}

############################################
# Common
############################################

# make distance matrices using shared features. Distance measures are from Korbel et al SHOT paper

sub make_qualitative_distance_matrices {
	use Smash::Utils::MatrixIO qw(write_phylip_matrix);
	my $feature_count = shift;
	my $replicate     = shift;
	my @labels   = sort {$a cmp $b} values %Labels; # has to be consistent across header/data
	my @features = keys %$feature_count;
	my %Count; # num features in each label
	my %Shared; # num features shared bw two labels
	foreach my $l1 (@labels) {
		$Count{$l1} = 0;
		foreach my $l2 (@labels) {
			$Shared{$l1}{$l2} = 0;
		}
	}
	foreach my $feature (@features) {
		my @positive_labels;
		foreach my $label (@labels) {
			if (defined($feature_count->{$feature}->{$label}) && $feature_count->{$feature}->{$label} == 1) {
				push(@positive_labels, $label);
				$Count{$label}++;
			}
		}
		foreach my $pos1 (@positive_labels) {
			foreach my $pos2 (@positive_labels) {
				$Shared{$pos1}{$pos2}++;
			}
		}
	}
	my %Dist;
	foreach my $l1 (@labels) {
		foreach my $l2 (@labels) {
			my ($c1, $c2) = ($Count{$l1}, $Count{$l2});
			my $normalization = $c1*$c2*sqrt(2 / ($c1*$c1 + $c2*$c2));
			my $sim = $Shared{$l1}{$l2}/$normalization;
			$Dist{$l1}{$l2} = -log($sim);
		}
	}
	my $distance_file = "$options{tag}.distance.shared.$options{analysis}.$refdb";
	write_phylip_matrix($distance_file, %Dist);
}

sub init_reporter {
	if ($options{mode} eq "itol") {
		my $desc = "$options{analysis} mapping (";
		if ($options{tag}) {
			$desc .= "Tag:$options{tag}; ";
		}
		if ($options{level}) {
			$desc .= "Level:$options{level}; ";
		}
		$desc .= sprintf("Database:%s)", join("_", @opt_refdb));

		my $smash = new Smash::Core();
		$smash->init();
		my %reporter_options = (
					    NAME => "$options{tag}.$options{analysis}",
					  LABELS => [map {$Labels{$_}} @entities],
				  DATASET_COLORS => \%Colors,
				    PICTURE_SIZE => 500,
				     DESCRIPTION => $desc,
				     GRAPH_WIDTH => 2.5,
				 UPLOAD_ID => $smash->get_conf_value("iTOL", "uploadID")
				);
		$smash->finish();

		if ($options{analysis} eq "functional") {
			$reporter_options{GRAPH_WIDTH} = 8;
			if ($options{level} eq "ko" || $options{level} eq "og") {
				$reporter_options{GRAPH_WIDTH} = 1;
			}
			$reporter_options{DISPLAY_THRESHOLD} = 0.0001;
			$reporter_options{NCBI_TAXONOMY_ID} = 0;
		} elsif ($options{analysis} eq "refgenome") {
			$reporter_options{DISPLAY_THRESHOLD} = 0.0001;
			$reporter_options{NCBI_TAXONOMY_ID} = 1;
			#$reporter_options{RE_SORT} = 1;
			$reporter_options{COLOR_CLADES}     = 1;
		} elsif ($options{analysis} eq "16S") {
			$reporter_options{DISPLAY_THRESHOLD} = 0.01;
			$reporter_options{NCBI_TAXONOMY_ID} = 1;
			$reporter_options{COLOR_CLADES}     = 1;
		}

		$reporter = new Smash::Utils::iTOL(%reporter_options);
		$reporter->init();
	}
}

sub finish_reporter {
	$reporter->finish() if $reporter;
}

sub get_analyzer {
	my $type = shift;
	my @available = qw(Functional RefGenome 16S);
	my ($name) = grep {lc($type) eq lc($_)} @available;
	return $analyzer = "Smash::Analyses::Comparative::$name";
}

__END__

=begin html

=head1 doComparativeAnalysis.pl

=head1 Name

doComparativeAnalysis.pl - Comparative analysis of metagenomes

=end html

=begin latex

=head1 Name

doComparativeAnalysis.pl - Comparative analysis of metagenomes

=end latex

=head1 Synopsis

	doComparativeAnalysis.pl [options]

=head1 Options

=over 4

=item B<C<--tag>> (required)

Label for this analysis. Every file created by this run of this step of the 
analysis will have the given label as its prefix, to help you perform a 
series of steps as an analysis unit.

=item B<C<--analysis>> (required)

Type of analysis you would like to perform. Currently available options are 
C<refgenome>, C<16S> and C<functional>.

=item B<C<--list>> (required)

Name of file containing a list of samples, one per line. For C<refgenome> and
C<16S> analysis, this should contain the metagenome ids. For C<functional>
analysis, this list should contain  the gene prediction ids, because there
can be multiple gene predictions per sample.

By default, this script assumes that the samples to be analyzed are in the 
database and will try to get additional information from the database. If you
want to analyze external data using SMASH, then you have to provide this
information through the list file (See L<Analysis types>) for more details).

=item B<C<--mode>> (required)

Which step of the analysis do you want to perform now? Available steps are
C<raw>, C<summary>, C<distmat>, C<bootstrap+distmat>, C<itol>.

=item B<C<--replicate>>

Replicate to process, if this is part of C<bootstrap> analysis. If this is not
specified, the whole dataset will be processed.

=item B<C<--level>>

Functional or phylogenetic level at which to perform the analysis. For 
C<refgenome> and C<16S>, these are phylogenetic levels. Please keep in mind that
some organisms do not have some taxonomic levels, so asking for that level will
assign the closest rank above the requested level available for that organism.
For C<functional> analysis, this must be one of C<og> (for orthologous groups),
C<module> (for KEGG modules, if the data contains KEGG annotations) or C<pathway>
(for KEGG pathways, if the data contains KEGG annotations).

=item B<C<--refdb>>

Name of the reference database used. Although it is recommended that you use the
real name of databases (e.g., eggnog if the db is the eggNOG proteins, NCBI if 
the db is the NCBI genomes), it is really used in identifying the file to parse.

For example, if file F<list.txt> contains

	MC20.MG1
	MC20.MG2
	MC30.MG5

and you call the script 

	doComparativeAnalysis.pl --analysis=refgenome --list=list.txt \
		--mode=raw --refdb=NCBI --tag=April01 ...

then the script expects F<MC20.MG1.NCBI.blastn>, F<MC20.MG2.NCBI.blastn> and 
F<MC30.MG5.NCBI.blastn> in the present working directory. This also creates 
an output file called F<April01.feature_vector.rawcount.B<refgenome>.B<NCBI>> 
(bold words are from the commandline options).

If file F<list.txt> contains

	MC20.MG1.AS2.GP2
	MC20.MG2.AS3.GP1
	MC30.MG5.AS1.GP3

and you call the script 

	doComparativeAnalysis.pl --analysis=functional --list=list.txt \
		--mode=raw --refdb=eggnog --tag=April01 ...

then the script expects F<MC20.MG1.AS2.GP2.eggnogmapping.txt>, F<MC20.MG2.AS3.GP1.eggnogmapping.txt> and 
F<MC30.MG5.AS1.GP3.eggnogmapping.txt> in the present working directory. This also creates 
an output file called F<April01.feature_vector.rawcount.B<functional>.B<eggnog>> 
(bold words are from the commandline options).

For historic reasons, this option can also take a comma-separated list of 
such C<refdb>'s to perform the analysis on multiple reference databases. 
For example, running the command

	doComparativeAnalysis.pl --analysis=functional --list=list.txt \
		--mode=rawcount --refdb=eggnog,kegg --tag=April01 ...

is equivalent to running the same command with C<--refdb=eggnog> and 
C<--refdb=kegg> sequentially. This was just an example, and unless you just have
one CPU to do all your analysis, we do not recommend running the command above.
Instead, run two analysis for C<--refdb=eggnog> and C<--refdb=kegg>
simultaneously so that you have your results in roughly half as much time.
However, this options comes in very handy when you use C<--mode=itol>. For
example, running the command

	doComparativeAnalysis.pl --analysis=refgenome --list=list.txt \
		--level=genus --mode=itol --refdb=NCBI,MyDB --tag=April01 ...

will parse the read mappings at the genus level using the NCBI database and
MyDB database, and upload them as two datasets to one underlying iTOL tree
which contains the genera hit in either database.

Unfortunately, this won't work for C<functional> analysis using C<kegg>
and C<eggnog>, since the orthologous groups have different names! The 
resulting tree will have all the eggNOG OGs and the KEGG KOs, and you
end up with a bizarre iTOL tree of mutually exclusive datasets that
shouldn't have uploaded together.

=item B<C<--normalize>>

Normalization procedure to be used. Available options are: C<data> that 
will normalize the abundances by the datasize, and C<hits> that will
normalize the abundances by the positive hits. For example, if a dataset
contains 1000 reads, 500 of which are mapped to some genus, and 50 of 
them are mapped to I<Lactobacillus>, then using C<hits> will result in
an abundance of 10% for I<Lactobacillus> (50 out of 500), whereas using 
C<data> will result in 5% (50 out of 1000). You get the idea!

=item B<C<--help>>

Prints this manual.

=back

=head2 Options specific to B<C<--mode=itol>>

=over 4

=item B<C<--top>>

Applies only to C<itol> mode. Use only the C<N> most abundant features, be
it genera, species or orthologous groups. This top C<N> filter is applied 
to each sample independently, so the final list could be more than C<N> long
when multiple samples have different sets of top C<N> features.

=item B<C<--filterlow>>

Applies only to C<itol> mode. Variant of C<--top>, it filters features (genera,
species, orthologous groups) that are below the given threshold in ALL samples.
If one sample is above the threshold, the feature remains in all samples.

=back

=head2 Options specific to B<C<--analysis=refgenome>>

=over 4

=item B<C<--identity>>

Percent identity cutoff to be used for this analysis. Only applies to 
C<refgenome> analysis where it is used as the threshold above which
a read is assigned to the organism of the BLAST hit at the taxonomic 
level specified by C<level>. Typical values we use are 65% for C<phylum>,
85% for C<genus> and 95% for C<species>.

=item B<C<--local>>

Applies only to C<refgenome> analysis. SMASH uses a database that maps each
sequence in the reference genome database to its source organism identified
by its NCBI taxonomy id. These sequences are also annotated for protein-coding
genes, non-coding genes, etc. If you do not have this database installed at 
your site, e.g. because you are calling this script from B<SmashCell>, you
probably have downloaded the information from SMASH website. Then this option
is just for you. Specify C<--local> together with C<--dbfile>.

=item B<C<--dbfile>>

This file contains information mapping each sequence in the reference genome
database to its source organism. You can download the latest version of this
file, along with the FASTA file containing the reference sequences from 
our downloads page.

If you have BLASTed your metagenome reads against that FASTA file, and your
output files are called F<MC20.MG1.May09.blastn>, F<MC20.MG2.May09.blastn> and
F<MC30.MG5.May09.blastn>, you can run the command

	doComparativeAnalysis.pl --analysis=refgenome --list=list.txt \
		--level=genus --mode=itol --refdb=May09 \
		--local --dbfile=reference_genomes_information.may09.txt \
		--tag=April01 ...

to upload this data to iTOL.

=item B<C<--genomesize | --nogenomesize>>

By default the script performs genome size normalization for the refgenome
analysis. If you do not need this, specify C<--nogenomesize> to turn this off.

=back

=head2 Options specific to B<C<--analysis=16S>>

=over 4

=item B<C<--conf_threshold>>

Minimum bootstrap confidence threshold reported by RDP
classifier. (default: 0.6)

=item B<C<--length_threshold>>

Minimum length of a read to be considered mapped to a rank. (default: 250)

=back

=head1 Description

C<doComparativeAnalysis.pl> is a multifunctional script that enables comparative
metagenomic analysis. Before I explain the functionalities of the script, I must
explain its dependencies on external information and/or resources.

=head2 External dependencies

=begin html

<BLOCKQUOTE>
<HR>

=end html

=head3 iTOL batch access

Many functionalities of this script generate cluster or trees and these can be 
uploaded to the iTOL webserver for viewing as well storage. SMASH uses the iTOL
batch access website through scripts kindly provided by iTOL. You must have an 
user account at the iTOL website to use this feature. When you upload a
tree through this script, the tree will be uploaded to a project called 
B<itol_uploader> under your user account. Please create this project under 
your user account, otherwise
the trees might be lost in digital space. Additionally, for iTOL to allow
batch access to your account, you must enable it. Please see 
L<http://itol.embl.de/help/batch_help.shtml> for details (under B<Enabling
batch upload>). Once you enable it, you receive an upload ID, which you should enter 
it in the SMASH config file under the B<iTOL> section as shown below:

	[Smash]
	data_dir          : /some/embl/location/smash/data_repos
	software_dir      : software
	workspace_dir     : workspace
	collection_prefix : MC
	metagenome_prefix : MG
	assembly_prefix   : AS
	genepred_prefix   : GP

	[SmashDB]
	database_engine : mysql
	database_name   : SmashDB
	user : smashdb
	pass : smashdb
	host : server
	port : 9999

	[RefOrganismDB]
	data_dir        : /some/embl/location/smash/data_repos/reference_organisms
	database_engine : sqlite3
	database_name   : /some/embl/location/smash/db/RefOrganismDB.sqlite
	user :
	pass : 
	host : 
	port : 

	[iTOL]
	uploadID : xyzw

=begin html

<HR>

=end html

=head3 NCBI Taxonomy information

Phylogenetic analysis performed by this script uses the taxonomy information from
NCBI. It uses a local repository of the NCBI taxonomy dump (normally available at
L<ftp://ftp.ncbi.nih.gov/pub/taxonomy>) to get detailed information on each 
organism. You must specify the remote and local repository location in the 
configuration file, in the B<Taxonomy> section as shown below:

	[Taxonomy]
	remote_repository : ftp://ftp.ncbi.nih.gov/pub/taxonomy
	local_repository  : /some/embl/location/smash/data_repos/external/taxonomy

When the script needs to parse this taxonomy information, it checks whether 
F<local_repository> contains the required files. If they do not exist, it downloads
them from F<remote_repository>. If you would like to update the files using the 
latest available from NCBI, you could replace the files in F<local_repository>
with the newer ones yourself, or update it automatically by running the script
L<updateNCBITaxonomyFiles.pl|updateNCBITaxonomyFiles> without any arguments:

	updateNCBITaxonomyFiles.pl

=begin html

<HR>

=end html

=head3 PHYLIP package

For clustering the samples, we use the C<neighbor> method from the B<phylip>
package. For generating consensus clusters from bootstrap analysis, we use the
C<consense> program from B<phylip> as well. Please make sure that C<neighbor>
and C<consense> are available in your path. You can get B<phylip> package from
L<http://evolution.genetics.washington.edu/phylip.html>.

=begin html

<HR>
</BLOCKQUOTE>

=end html

=head2 Analysis types

Currently, three types of analysis are possible - two phylogenetic and one
functional analyses.

=head3 Phylogenetic analysis using reference genome mapping

This is the most sensitive phylogenetic analysis for a metagenome sample provided
of course that the species population has relatives in the reference genome set.
The following are the prerequisites before running this analysis:

=over 4

=over 4

=item 1. Sample information

=over 4

=item a. The metagenome exists in SMASH. 

(e.g., MC20.MG1) and reads have been loaded into SMASH using 
L<loadMetaGenome.pl|loadMetaGenome>. In this case, the 
script will retrieve the number of reads, sample labels etc from the database.

=item (or)

=item b. You give this through the list file using the C<--list> option.

For example, F<list.txt> could contain sample names and sample size (in terms of 
read counts) like this:

	SampleA     80000
	SampleB     76000
	SampleC_T1  97000
	SampleC_T2  98000

If it looks like this:

	SampleA     80000  S_A
	SampleB     76000  S_B
	SampleC_T1  97000  S_C1
	SampleC_T2  98000  S_C2

then S_A will be used to denote SampleA, S_B for SampleB and so on in the final output
files.

=back

=item 2. You have a set of reference genome sequences. 

These could be WGS scaffolds, 
contigs, or complete genomes. B<Please do not use a set of WGS reads since this 
will mess up genome size calculations.> Let us call this set B<May09>.

=item 3. You have the taxonomical information of these. 

Either these are available 
in the C<RefOrganismDB> database in SMASH, or you have a flat file downloaded from 
L<http://www.bork.embl.de/software/smash/downloads/reference_genomes/>. If you make
your own reference genome database, you could make this file yourself: please check
the header (first line) of the file for which fields contain what information.

=item 4. Reads have been BLASTed against the reference genome sequences.  

The output of this BLAST run must be F<MC20.MG1.May09.blastn>.

=back

=back

We are now ready to run the comparative phylogenetic analysis. The first step is to
parse the BLAST outputs.

=over

=item B<1. Parse the BLAST output:>

You can now run

	doComparativeAnalysis.pl --analysis=refgenome --list=list.txt \
		--local --dbfile=reference_genomes_information.may09.txt \
		--mode=raw --refdb=May09 --tag=trial1 \
		--normalize=data --identity=85 \
	> read_refgenome_map.txt

This will produce the file F<trial1.feature_vector.rawcount.refgenome.May09>. This 
file is in C<R> table format, containing NCBI taxonomy ids as rows and samples (in 
this case just MC20.MG1) as columns. It could look like:

	MC20.MG1	MC20.MG2	MC30.MG5
	-1	489	387	899
	435590	1338	994	1232
	272559	839	789	403
	1000537011	241	358	491

which means that, for sample MC20.MG1, 23% of the reads could not be mapped 
at all (that's what -1 means),
41% of the reads were mapped to I<Bacteroides vulgatus ATCC 8482> (taxonomy id 
435590), 22% of the reads to I<Bacteroides fragilis NCTC 9343> (taxonomy id 272559)
 and 14% of the reads were mapped to I<Prevotella copri DSM 18205> 
(taxonomy id 537011), all above the sequence identity threshold of 85%. 

Once we have parsed the BLAST files and created the file above, there are several
things we could do.

=item B<2. Cluster samples without bootstrapping>

=over 4

=item B<2a. Estimate inter-sample distances:>

Since this threshold is suitable for genus level mapping, we will now create 
the genus abundance of this sample.

	doComparativeAnalysis.pl --analysis=refgenome --list=list.txt \
		--local --dbfile=reference_genomes_information.may09.txt \
		--mode=distmat --refdb=May09 --tag=trial1 \
		--normalize=data --identity=85 --level=genus

This will generate a normalized genus level abundance profile like this:

	MC20.MG1	MC20.MG2	MC30.MG5
	-1	0.00013818	0.00010936	0.00025404
	816	0.00042033	0.00034410	0.00031604
	838	0.00007104	0.00010553	0.00014473

in the file F<trial1.feature_vector.rawcount.refgenome.genus.May09>.

These numbers are generated after two steps. First the number of reads mapping to
each genome is normalized by the genome length. Then these normalized values are
added  according to the genus, 
where 816 and 838 are the NCBI taxonomy ids of I<Bacteroides> and I<Prevotella>, 
respectively. The abundances of all genomes belonging to I<Bacteroides> have been
summed together into the abundance of I<Bacteroides>. This resides in a file
F<trial1.feature_vector.rawcount.refgenome.genus.May09>. This profile is used to
generate a normalized genus level relative abundance distribution like this:

	MC20.MG1	MC20.MG2	MC30.MG5
	-1	0.21949439	0.19564004	0.35539844
	Bacteroides	0.66766555	0.61557832	0.44212809
	Prevotella	0.11284007	0.18878164	0.20247347

in the file F<trial1.feature_vector.jsd.refgenome.genus.May09>.

This abundance distribution is then used to make a distance matrix in phylip format:

	3
	MC20.MG1   0.000000 0.075521 0.161573
	MC20.MG2   0.075521 0.000000 0.137878
	MC30.MG5   0.161573 0.137878 0.000000

in the file F<trial1.distance.jsd.refgenome.genus.May09>.

=item B<2b. Cluster using these distances:>

If you want to cluster these samples using this distance matrix, then run this:

	doComparativeAnalysis.pl --analysis=refgenome --list=list.txt \
		--mode=cluster --refdb=May09 --tag=trial1 \
		--level=genus --identity=85 --normalize=data \
		--local --dbfile=reference_genomes_information.may09.txt

This will use C<neighbor> method from the C<phylip> package to cluster these
samples, and upload this tree to the iTOL website.

=back

=item B<3. Cluster samples with bootstrapping>

=over 4

=item B<3a. Estimate inter-sample distances:>

If you run, instead

	doComparativeAnalysis.pl --analysis=refgenome --list=list.txt \
		--mode=bootstrap+distmat --refdb=May09 --tag=trial1 \
		--normalize=data --level=genus

this will make 100 bootstrap replicates from the input file, and create a 
distance matrix as mentioned above for each replicate.

=item B<3b. Cluster using these distance:>

If you want to cluster these samples with bootstrap support, then run this:

	doComparativeAnalysis.pl --analysis=refgenome --list=list.txt \
		--mode=cluster --refdb=May09 --tag=trial1 \
		--level=genus --identity=85 --normalize=data \
		--local --dbfile=reference_genomes_information.may09.txt

This will read all the 100 distance matrices, use the C<neighbor> program to
make 100 trees and then use the C<consense> program to produce a consensus 
tree with bootstrap support, which will be loaded to iTOL.

=back

=item B<4. Display the phylogenetic composition on iTOL>

If you would like to display the genus distribution as graphs using iTOL, you
can upload this to the iTOL server by running the following command:

	doComparativeAnalysis.pl --analysis=refgenome --list=list.txt \
		--mode=itol --refdb=May09 --tag=trial1 \
		--level=genus --identity=85 --normalize=data \
		--local --dbfile=reference_genomes_information.may09.txt

A successful run looks like this:

	Avg. genome size = 3538803
	Processing samples ... done
	Filtering samples ... done
	Running: /usr/bin/perl /usr/local/Smash/bin/iTOL_uploader.pl --config trial1.refgenome.cfg
	ITOL: 
	ITOL: iTOL batch uploader
	ITOL: ===================
	ITOL: Upload successful. Your tree is accessible using the following iTOL tree ID:
	ITOL: 
	ITOL: 1011041672954712702391540
	ITOL: 
	Upload successful. You can view this tree on iTOL using this URL:
	http://itol.embl.de/external.cgi?tree=1011041672954712702391540
	Running: /usr/bin/perl /usr/local/Smash/bin/iTOL_downloader.pl --tree 10110416729547127023
	91540 --format svg --outputFile trial1.refgenome.downloaded.svg --fontSize 20 --displayMod
	e normal --colorBranches 1 --alignLabels 1 --scaleFactor 1 --showBS 1 --BSdisplayValue M0 
	--BSdisplayType text --datasetList dataset1

	iTOL batch downloader
	=====================
	Exported tree saved to trial1.refgenome.downloaded.svg

The script tells you the commands it is executing to upload and download the 
tree to iTOL. Lines beginning with C<"ITOL:"> are what it received as a 
response from iTOL webserver, for your information. When the upload is successful, 
you can view the tree following the link provided. 
Hovering over any bar will show the name of the dataset, and the abundance
of the corresponding genus in that dataset.

=begin html

<IMG ALIGN="center" WIDTH="50%" SRC="../images/smash_tree_on_itol.png" HREF="../images/smash_tree_on_itol.png" ALT="Phylogenetic compositions on iTOL">

=end html


You can also use the SVG 
file listed in the last line for the tree that contains additional 
label information. 

=begin html

<IMG ALIGN="center" WIDTH="70%" SRC="../images/smash_tree_postprocessed.png" HREF="../images/smash_tree_postprocessed.png" ALT="Phylogenetic compositions downloaded from iTOL">

=end html


=back

=head1 Dependencies

doComparativeAnalysis.pl requires the following PERL modules:

	Config, Pod::Usage 

The iTOL batch access scripts require the following PERL modules:

	Getopt::Regex, Config::File::Simple, HTTP::Request::Common and LWP::UserAgent.

=head1 Author

Mani Arumugam (arumugam@embl.de, http://www.embl.de)

=head1 Copyright

Copyright (C) 2009-2010. Mani Arumugam and EMBL. All rights reserved.

=cut
