#! /usr/bin/env perl
use strict;
use warnings;
use Smash::Utils::ClusterBLAST;
use Getopt::Long;

my $opt_blast;
my $opt_blastdb;
my $opt_query;
my $opt_label;
my $opt_bits = 60;
my $opt_E = 0.1;
my $opt_outdir;
my $opt_tabular = 1;
my $opt_param;
my $opt_cpus = 1;
my $opt_pieces;
my $opt_max_subjects = 10000;
my $opt_stage = "tmp";

my $opt_help;

my $prologue = "";

my $usage = "
$0 - Run WU-BLAST on clusters.
Splits a query file into small files, submits to the cluster and combines the results.

Usage: $0 [options]

Required parameters:
 --blast=<program>     BLAST program (blastn|blastp|blastx|tblastn|tblastx)
 --db=<prefix>         name of the blast database file
 --qry=<file>          query file
 --label=<name>        name for the job on cluster
 --outdir=<dir>        directory to store the output file
 --pieces=<num>        (approximate) number of pieces to split the query file into 

Output file will be stored as <out_dir>/<label>.<program>

Optional Parameters:

BLAST options:
 --E=<num>             E-value threshold for results to be included, passed through -E to blast (default: $opt_E)
 --maxS=<num>          maximum number of subjects per query to report, passed through -B to blast (default: $opt_max_subjects)
 --(no)tabular         whether to use tabular output of WU-BLAST (default: true)
                        notabular only works when --pieces=1
 --param=<string>      additional parameters passed to blast as they are. Use quotes if you use multiple parameters

Cluster parallelization options:
 --cpus=<num>          number of cpus to use for every subjob (default: $opt_cpus)
 --stage=<mode>        local staging location (memory or tmp) in the node to copy the blast database (default: $opt_stage)

Post processing Options:
 --bits=<num>          bit threshold for results to be included (default: $opt_bits)

";

GetOptions(
	"blast=s"   => \$opt_blast,
	"db=s"      => \$opt_blastdb,
	"qry=s"     => \$opt_query,
	"label=s"   => \$opt_label,
	"bits=n"    => \$opt_bits,
	"E=f"       => \$opt_E,
	"outdir=s"  => \$opt_outdir,
	"tabular!"  => \$opt_tabular,
	"param=s"   => \$opt_param,
	"cpus=n"    => \$opt_cpus,
	"pieces=n"  => \$opt_pieces,
	"maxS=n"    => \$opt_max_subjects,
	"stage=s"   => \$opt_stage,
	"help"      => \$opt_help);

($opt_blast && $opt_blastdb && $opt_query && $opt_label && $opt_outdir && $opt_param && $opt_pieces && !$opt_help) || die $usage;


# if you want it ultrafast, put the database in shared memory

my $node_location;
if ($opt_stage eq "memory") {
	$node_location = "/dev/shm/".getpwuid($<)."/in";
} else {
	if ($opt_stage ne "tmp") {
		warn "Stage option $opt_stage not recognized. Using 'tmp' instead!\n";
	}
	$node_location = "/tmp/".getpwuid($<)."/in";
}

Smash::Utils::ClusterBLAST->run_blast_cluster(  BLAST      => $opt_blast,
						BLASTDB    => $opt_blastdb,
						QUERY      => $opt_query,
						LABEL      => $opt_label,
						BITS       => $opt_bits,
						E          => $opt_E,
						OUTDIR     => $opt_outdir,
						TABULAR    => $opt_tabular,
						PARAMETERS => $opt_param,
						CPUS       => $opt_cpus,
						PIECES     => $opt_pieces,
						STARTONLY  => 1,
						NODETEMP   => $node_location,
						PROLOGUE   => $prologue,
						MAXSUBJECTS=> $opt_max_subjects);
