#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "mDNA.h"

#define COLUMNS 60

void mFreeDNA(mDNA *dna) {
	mFree(dna->def);
	mFree(dna->data);
	mFree(dna->code);
}

void mFreeDNAVector(mVector *vec) {
	int i;
	for (i=0; i<vec->size; i++) {
		mFreeDNA((mDNA*)vec->elem[i]);
		mFree(vec->elem[i]);
	}
	mFreeVector(vec);
}

/* returns 1 - if there is no more fasta entry left,
 *         2 - if there is more fasta entries left
 */
int mReadDNA(FILE *stream, mDNA *dna) {
	int     next = 1;
	char   *data;
	int    *code;
	char    def[256];
	int     c;
	int     gc = 0, at = 0;
	coor_t  count = 0;
	coor_t  limit = 512; /* initial mem alloc */

	data = (char*) mMalloc(limit*sizeof(char));
	code = (int*)  mMalloc(limit*sizeof(int));

	while (isspace(c = fgetc(stream)));
	if ('>' == (char) c) { /* Header line found */
		while ( (c = fgetc(stream)) != EOF && c != '\n'  && count < 256) { 
			/* Process header */
			def[count++] = (char) c;
		}
		def[count] = '\0';
	} else {
		/*mExit("Not a valid Fasta file!");*/
		mDie("Not a valid Fasta file!");
	}

	count = 0;
	for(;;){
		c = fgetc(stream);
		if (c == EOF || c == '>') {
			/* next record */
			break;
		}
		if (isalpha(c)) {
			if (count == limit) {
				limit *= 2;
				data = (char*) mRealloc(data, limit*sizeof(char));
				code = (int*)  mRealloc(code, limit*sizeof(int));
			}
			data[count] = (char) c;
			switch(c) {
				case 'a':
				case 'A':
					code[count] = 0; at++;
					break;
				case 'c':
				case 'C':
					code[count] = 1; gc++;
					break;
				case 'g':
				case 'G':
					code[count] = 2; gc++;
					break;
				case 't':
				case 'T':
					code[count] = 3; at++;
					break;
				case 'n':
				case 'N':
				default:
					code[count] = 4;
					break;
			}
			count++;
		}
	}
	if (c == '>') {
		/* next record */
		ungetc(c, stream);
		next = 2;
	}

	dna->length = count;
	dna->def = (char*) mMalloc((strlen(def)+1)*sizeof(char));
	strcpy(dna->def, def);
	dna->data = data;
	dna->code = code;
	dna->gc   = 1.0 * gc / (gc+at);
	return next;
}

int mReadMultipleDNA(FILE *stream, mVector *multi) {
	int status;
	mDNA*  dna   = (mDNA*)  mMalloc(sizeof(mDNA));
	while((status = mReadDNA(stream, dna))) {
		mPushVector(multi, dna);
		if (status == 1) { /* Last entry */
			break;
		}
		dna = (mDNA*) mMalloc(sizeof(mDNA));
	}
	return multi->size;
}

void mWriteMultipleDNA(FILE *stream, mVector *multi) {
	int i;
	for (i=0; i<multi->size; i++) {
		mWriteDNA(stream, (mDNA*)multi->elem[i]);
	}
}

void mWriteDNA(FILE *stream, mDNA *dna) {
	int i;
	int   length = dna->length;
	char *data   = dna->data;
	fprintf(stream, ">%s\n", dna->def);
	for (i=0; i < length; i++) {
		fprintf(stream, "%c", data[i]);
		if (i%COLUMNS == COLUMNS-1) {
			fprintf(stream, "\n");
		}
	}
	if (length%COLUMNS != 0) fprintf(stream, "\n");
}

void mComplementDNA(mDNA *dna) {
	int i;
	int length = dna->length;
	char *data = dna->data;
	for (i=0; i<length; i++) {
		switch(data[i]) {
			case 'A':
				data[i] = 'T';
				break;
			case 'a':
				data[i] = 't';
				break;
			case 'C':
				data[i] = 'G';
				break;
			case 'c':
				data[i] = 'g';
				break;
			case 'G':
				data[i] = 'C';
				break;
			case 'g':
				data[i] = 'c';
				break;
			case 'T':
				data[i] = 'A';
				break;
			case 't':
				data[i] = 'a';
				break;
		}
		/* weird error: for 'N', this becomes -1 and not 4.
		 * so I added a 5 in front just in case */
		dna->code[i] = (5 + 3 - dna->code[i])%5; 
	}
	dna->gc = 1 - dna->gc;
}

void mReverseDNA(mDNA *dna) {
	int i;
	int length = dna->length;
	int *code = (int*) mMalloc(sizeof(int)*length);
	char *data = (char*) mMalloc(sizeof(char)*length);
	for (i=0; i<length; i++) {
		code[i] = dna->code[length-i-1];
		data[i] = dna->data[length-i-1];
	}
	for (i=0; i<length; i++) {
		dna->code[i] = code[i];
		dna->data[i] = data[i];
	}
	mFree(code);
	mFree(data);
}

void mReverseComplementDNA(mDNA *dna) {
	mReverseDNA(dna);
	mComplementDNA(dna);
}

char* mDNA2Char(mDNA *dna) {
	char *seq = (char*) mMalloc(sizeof(char)*(dna->length+1));
	memcpy(seq, dna->data, dna->length*sizeof(char));
	seq[dna->length] = '\0';
	return seq;
}

void mProcessDNADef(mDNA *dna) {
	char *def    = dna->def;
	int   length = strlen(def);
	int   i;
	for (i=0; i<length; i++) {
		if (isspace(def[i])) {
			def[i] = '\0';
		}
	}
}

mDNA* mSubDNA(mDNA *dna, coor_t offset, coor_t length) {
	mDNA *sub;

	if (offset >= dna->length) return NULL;

	sub         = (mDNA*) mMalloc(sizeof(mDNA));
	sub->data   = dna->data + (int)offset;
	sub->code   = dna->code + (int)offset;
	sub->def    = dna->def;
	sub->length = length;
	if (offset + length > dna->length) {
		sub->length = dna->length - offset;
	}

	/* Don't calculate GC percentage. A call to mGetGC will calculate it. */
	sub->gc = -FLT_MAX;
	return sub;
}

float mGetGC(mDNA *dna) {
	int i, at=0, gc=0;
	int  *code;
	if (dna->gc != -FLT_MAX) {
		return dna->gc;
	}
	/* Calculate GC percentage */
	code = dna->code;
	for (i=0; i<dna->length; i++) {
		switch(code[i]) {
			case 0: at++; break;
			case 1: gc++; break;
			case 2: gc++; break;
			case 3: at++; break;
		}
	}
	dna->gc = 1.0 * gc/(gc+at);
	return dna->gc;
}

void mBreakAndWriteDNA(FILE *stream, mDNA *dna, coor_t chunk) {
	char *def, *dup, *brk;
	coor_t i;
	int count;

	def  = (char*) mMalloc((strlen(dna->def)+2*((int)log10(dna->length/chunk)) + 3)*sizeof(char));
	dup  = (char*) mMalloc((strlen(dna->def) + 1)*sizeof(char));
	strcpy(dup, dna->def);
	brk = strpbrk(dup, " \t");
	if (brk != NULL) {
		*brk = '\0';
	}
	for (i=0; i<dna->length; i+=chunk) {
		mDNA *sub = mSubDNA(dna, i, chunk);
		sprintf(def, "%s:%ld-%ld", dup, i, i+sub->length-1);
		sub->def  = def;
		mWriteDNA(stream, sub);
		count++;
	}
	mFree(def);
	mFree(dup);
}

void mSplitDNAToFiles(char *prefix, mDNA *dna, coor_t chunk) {
	FILE *stream;
	char *def, *dup, *brk, *filename;
	coor_t i;
	int count;

	filename  = (char*) mMalloc((strlen(prefix)+((int)log10(dna->length/chunk)+1) + 2)*sizeof(char));
	def  = (char*) mMalloc((strlen(dna->def)+2*((int)log10(dna->length/chunk)) + 3)*sizeof(char));
	dup  = (char*) mMalloc((strlen(dna->def) + 1)*sizeof(char));
	strcpy(dup, dna->def);
	brk = strpbrk(dup, " \t");
	if (brk != NULL) {
		*brk = '\0';
	}
	count = 0;
	for (i=0; i<dna->length; i+=chunk) {
		mDNA *sub = mSubDNA(dna, i, chunk);
		sprintf(def, "%s:%ld-%ld", dup, i, i+sub->length-1);
		sprintf(filename, "%s.%d", prefix, count);
		sub->def  = def;
		if ((stream = fopen(filename, "w")) == NULL) {
			mDie("Cannot open %s for writing", filename);
		}
		mWriteDNA(stream, sub);
		fclose(stream);
		count++;
	}
	mFree(def);
	mFree(dup);
	mFree(filename);
}
